from pathlib import Path
from typing import Optional

import click
import typer
from typing_extensions import Annotated

from janus.cli.constants import REFINERS
from janus.language.naive.registry import CUSTOM_SPLITTERS
from janus.utils.enums import LANGUAGES


def partition(
    input_dir: Annotated[
        Path,
        typer.Option(
            "--input",
            "-i",
            help="The directory containing the source code to be partitioned. ",
        ),
    ],
    language: Annotated[
        str,
        typer.Option(
            "--language",
            "-l",
            help="The language of the source code.",
            click_type=click.Choice(sorted(LANGUAGES)),
        ),
    ],
    output_dir: Annotated[
        Path,
        typer.Option(
            "--output-dir", "-o", help="The directory to store the partitioned code in."
        ),
    ],
    llm_name: Annotated[
        str,
        typer.Option(
            "--llm",
            "-L",
            help="The custom name of the model set with 'janus llm add'.",
        ),
    ] = "gpt-4o",
    failure_dir: Annotated[
        Optional[Path],
        typer.Option(
            "--failure-directory",
            "-f",
            help="The directory to store failure files during translation",
        ),
    ] = None,
    max_prompts: Annotated[
        int,
        typer.Option(
            "--max-prompts",
            "-m",
            help="The maximum number of times to prompt a model on one functional block "
            "before exiting the application. This is to prevent wasting too much money.",
        ),
    ] = 10,
    overwrite: Annotated[
        bool,
        typer.Option(
            "--overwrite/--preserve",
            help="Whether to overwrite existing files in the output directory",
        ),
    ] = False,
    temperature: Annotated[
        float,
        typer.Option("--temperature", "-t", help="Sampling temperature.", min=0, max=2),
    ] = 0.7,
    splitter_type: Annotated[
        str,
        typer.Option(
            "-S",
            "--splitter",
            help="Name of custom splitter to use",
            click_type=click.Choice(list(CUSTOM_SPLITTERS.keys())),
        ),
    ] = "file",
    refiner_types: Annotated[
        list[str],
        typer.Option(
            "-r",
            "--refiner",
            help="List of refiner types to use. Add -r for each refiner to use in\
                refinement chain",
            click_type=click.Choice(list(REFINERS.keys())),
        ),
    ] = ["JanusRefiner"],
    max_tokens: Annotated[
        int,
        typer.Option(
            "--max-tokens",
            "-M",
            help="The maximum number of tokens the model will take in. "
            "If unspecificed, model's default max will be used.",
        ),
    ] = None,
    partition_token_limit: Annotated[
        int,
        typer.Option(
            "--partition-tokens",
            "-pt",
            help="The limit on the number of tokens per partition.",
        ),
    ] = 8192,
):
    from janus.converter.partition import Partitioner

    refiner_types = [REFINERS[r] for r in refiner_types]
    model_arguments = dict(temperature=temperature)
    kwargs = dict(
        model=llm_name,
        model_arguments=model_arguments,
        source_language=language,
        max_prompts=max_prompts,
        max_tokens=max_tokens,
        splitter_type=splitter_type,
        refiner_types=refiner_types,
        partition_token_limit=partition_token_limit,
    )
    partitioner = Partitioner(**kwargs)
    partitioner.translate(input_dir, output_dir, failure_dir, overwrite)



from langchain.schema.output_parser import BaseOutputParser
from langchain_core.exceptions import OutputParserException
from langchain_core.messages import BaseMessage
from langchain_core.output_parsers import StrOutputParser

from janus.language.block import CodeBlock
from janus.language.splitter import EmptyTreeError
from janus.utils.logger import create_logger

log = create_logger(__name__)


class JanusParser(BaseOutputParser[str]):
    def parse_input(self, block: CodeBlock) -> str:
        """Parse the input block into raw string input ready to be passed to
        an LLM. Also perform any processing or saving of metadata.

        Arguments:
            block: The CodeBlock to be processed

        Returns:
            A parsed version of the input text
        """
        if block.text is None:
            raise EmptyTreeError("No text in input CodeBlock!")
        return block.text

    def parse_combined_output(self, text: str) -> str:
        """Parse the output text from the LLM when multiple inputs are combined

        Arguments:
            text: The output text from the LLM

        Returns:
            A parsed version of the text
        """
        return text

    def parse_into_block(self, text: str | BaseMessage, block: CodeBlock):
        if isinstance(text, BaseMessage):
            text = str(text.content)
        block.text = text


class GenericParser(JanusParser, StrOutputParser):
    def parse(self, text: str | BaseMessage) -> str:
        if isinstance(text, BaseMessage):
            text = str(text.content)
        return text

    def get_format_instructions(self) -> str:
        return "Output should be a string"


class JanusParserException(OutputParserException):
    def __init__(self, unparsed_output, *args, **kwargs):
        self.unparsed_output = unparsed_output
        super().__init__(*args, **kwargs)



import os
import time
from pathlib import Path
from typing import List, Optional

import joblib
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain_core.documents import Document
from langchain_unstructured import UnstructuredLoader
from sklearn.feature_extraction.text import CountVectorizer, TfidfVectorizer
from sklearn.metrics.pairwise import cosine_similarity

from janus.utils.logger import create_logger

log = create_logger(__name__)


class PDFDocsReader:
    def __init__(
        self,
        language: str,
        chunk_size: int = 1000,
        chunk_overlap: int = 100,
        start_page: Optional[int] = None,
        end_page: Optional[int] = None,
        vectorizer: CountVectorizer = TfidfVectorizer(),
    ):
        self.retrieval_docs_dir: Path = Path(
            os.getenv("RETRIEVAL_DOCS_DIR", "retrieval_docs")
        )
        self.language = language
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.start_page = start_page
        self.end_page = end_page
        self.vectorizer = vectorizer
        self.documents = self.load_and_chunk_pdf()
        self.doc_vectors = self.vectorize_documents()

    def load_and_chunk_pdf(self) -> List[str]:
        pdf_path = self.retrieval_docs_dir / f"{self.language}.pdf"
        pickled_documents_path = (
            self.retrieval_docs_dir / f"{self.language}_documents.pkl"
        )

        if pickled_documents_path.exists():
            log.debug(
                f"Loading pre-chunked PDF from {pickled_documents_path}. "
                f"If you want to regenerate retrieval docs for {self.language}, "
                f"delete the file at {pickled_documents_path}, "
                f"then add a new {self.language}.pdf."
            )
            documents = joblib.load(pickled_documents_path)
        else:
            if not pdf_path.exists():
                raise FileNotFoundError(
                    f"Language docs retrieval is enabled, but no PDF for language "
                    f"'{self.language}' was found. Move a "
                    f"{self.language} reference manual to "
                    f"{pdf_path.absolute()} "
                    f"(the path to the directory of PDF docs can be "
                    f"set with the env variable 'RETRIEVAL_DOCS_DIR')."
                )
            log.info(
                f"Chunking reference PDF for {self.language} using unstructured - "
                f"if your PDF has many pages, this could take a while..."
            )
            start_time = time.time()
            loader = UnstructuredLoader(
                pdf_path,
                chunking_strategy="basic",
                max_characters=1000000,
                include_orig_elements=False,
                start_page=self.start_page,
                end_page=self.end_page,
            )
            docs = loader.load()
            text = "\n\n".join([doc.page_content for doc in docs])
            text_splitter = RecursiveCharacterTextSplitter(
                chunk_size=self.chunk_size, chunk_overlap=self.chunk_overlap
            )
            documents = text_splitter.split_text(text)
            log.info(f"Document store created for language: {self.language}")
            end_time = time.time()
            log.info(
                f"Processing time for {self.language} PDF: "
                f"{end_time - start_time} seconds"
            )

            joblib.dump(documents, pickled_documents_path)
            log.debug(f"Documents saved to {pickled_documents_path}")

        return documents

    def vectorize_documents(self) -> (TfidfVectorizer, any):
        doc_vectors = self.vectorizer.fit_transform(self.documents)
        return doc_vectors

    def search_language_reference(
        self,
        query: List[str],
        top_k: int = 1,
        min_similarity: float = 0.1,
    ) -> List[Document]:
        """Searches through the vectorized PDF for the query using
        tf-idf and returns a list of langchain Documents."""

        docs: List[Document] = []

        for item in query:
            # Transform the query using the TF-IDF vectorizer
            query_vector = self.vectorizer.transform([item])

            # Calculate cosine similarities between the query and document vectors
            similarities = cosine_similarity(query_vector, self.doc_vectors).flatten()

            # Get the indices of documents with similarity above the threshold
            valid_indices = [
                i for i, sim in enumerate(similarities) if sim >= min_similarity
            ]

            # Sort the valid indices by similarity score in descending order
            sorted_indices = sorted(
                valid_indices, key=lambda i: similarities[i], reverse=True
            )

            # Limit to top-k results
            top_indices = sorted_indices[:top_k]

            # Retrieve the top-k most relevant documents
            docs += [Document(page_content=self.documents[i]) for i in top_indices]
            log.debug(f"Langauge documentation search result: {docs}")

        return docs



import json
import random
import uuid
from typing import List

from langchain.output_parsers import PydanticOutputParser
from langchain_core.exceptions import OutputParserException
from langchain_core.messages import BaseMessage
from langchain_core.pydantic_v1 import BaseModel, Field, validator

from janus.language.block import CodeBlock
from janus.parsers.parser import JanusParser
from janus.utils.logger import create_logger

log = create_logger(__name__)
RNG = random.Random()


class Criteria(BaseModel):
    reasoning: str = Field(description="A short explanation for the given assessment")
    score: str = Field("A simple `pass` or `fail`")

    @validator("score")
    def score_is_valid(cls, v: str):
        v = v.lower().strip()
        if v not in {"pass", "fail"}:
            raise OutputParserException("Score must be either 'pass' or 'fail'")
        return v


class Requirement(BaseModel):
    requirement_id: str = Field(description="The 8-character comment ID")
    requirement: str = Field(description="The original requirement being evaluated")
    C1: Criteria
    C2: Criteria
    C3: Criteria
    C4: Criteria
    C5: Criteria
    C6: Criteria
    C7: Criteria
    C8: Criteria
    C9: Criteria


class RequirementList(BaseModel):
    __root__: List[Requirement] = Field(
        description=(
            "A list of requirement evaluations. Each element should include"
            " the requirement's 8-character ID in the `requirement_id` field,"
            " the original requirement in the 'requirement' field, "
            " and nine score objects corresponding to each criterion."
        )
    )


class IncoseParser(JanusParser, PydanticOutputParser):
    requirements: dict[str, str]

    def __init__(self):
        PydanticOutputParser.__init__(
            self,
            pydantic_object=RequirementList,
            requirements={},
        )

    def parse_input(self, block: CodeBlock) -> str:
        # TODO: Perform comment stripping/placeholding here rather than in script
        text = super().parse_input(block)
        RNG.seed(text)

        obj = json.loads(text)

        # For some reason requirements objects are in a double list?
        reqs = obj["requirements"]

        # Generate a unique ID for each requirement (ensure they are unique)
        req_ids = set()
        while len(req_ids) < len(reqs):
            req_ids.add(str(uuid.UUID(int=RNG.getrandbits(128), version=4))[:8])

        self.requirements = dict(zip(req_ids, reqs))
        reqs_str = "\n\n".join(
            f"Requirement {rid} : {req}" for rid, req in self.requirements.items()
        )
        obj["requirements"] = reqs_str
        return json.dumps(obj)

    def parse(self, text: str | BaseMessage) -> str:
        if isinstance(text, BaseMessage):
            text = str(text.content)

        # Strip everything outside the JSON object
        begin, end = text.find("["), text.rfind("]")
        text = text[begin : end + 1]

        try:
            out: RequirementList = super().parse(text)
        except json.JSONDecodeError as e:
            log.debug(f"Invalid JSON object. Output:\n{text}")
            raise OutputParserException(f"Got invalid JSON object. Error: {e}")

        evals: dict[str, dict] = {c.requirement_id: c.dict() for c in out.__root__}

        seen_keys = set(evals.keys())
        expected_keys = set(self.requirements.keys())
        missing_keys = expected_keys.difference(seen_keys)
        invalid_keys = seen_keys.difference(expected_keys)
        if missing_keys:
            log.debug(f"Missing keys: {missing_keys}")
            if invalid_keys:
                log.debug(f"Invalid keys: {invalid_keys}")
            log.debug(f"Missing keys: {missing_keys}")
            raise OutputParserException(
                f"Got invalid return object. Missing the following expected "
                f"keys: {missing_keys}"
            )

        for key in invalid_keys:
            del evals[key]

        for rid in evals.keys():
            evals[rid]["requirement"] = self.requirements[rid]
            evals[rid].pop("requirement_id")

        return json.dumps(evals)

    def parse_combined_output(self, text: str) -> str:
        if not text.strip():
            return str({})
        objs = [json.loads(line.strip()) for line in text.split("\n") if line.strip()]
        output_obj = {}
        for obj in objs:
            output_obj.update(obj)
        return json.dumps(output_obj)



from pathlib import Path
from typing import Optional

import click
import typer
from typing_extensions import Annotated

from janus.cli.constants import REFINERS
from janus.language.naive.registry import CUSTOM_SPLITTERS
from janus.utils.enums import LANGUAGES


def diagram(
    input_dir: Annotated[
        Path,
        typer.Option(
            "--input",
            "-i",
            help="The directory containing the source code to be translated. "
            "The files should all be in one flat directory.",
        ),
    ],
    language: Annotated[
        str,
        typer.Option(
            "--language",
            "-l",
            help="The language of the source code.",
            click_type=click.Choice(sorted(LANGUAGES)),
        ),
    ],
    output_dir: Annotated[
        Path,
        typer.Option(
            "--output-dir", "-o", help="The directory to store the translated code in."
        ),
    ],
    llm_name: Annotated[
        str,
        typer.Option(
            "--llm",
            "-L",
            help="The custom name of the model set with 'janus llm add'.",
        ),
    ],
    failure_dir: Annotated[
        Optional[Path],
        typer.Option(
            "--failure-directory",
            "-f",
            help="The directory to store failure files during translation",
        ),
    ] = None,
    max_prompts: Annotated[
        int,
        typer.Option(
            "--max-prompts",
            "-m",
            help="The maximum number of times to prompt a model on one functional block "
            "before exiting the application. This is to prevent wasting too much money.",
        ),
    ] = 10,
    overwrite: Annotated[
        bool,
        typer.Option(
            "--overwrite/--preserve",
            help="Whether to overwrite existing files in the output directory",
        ),
    ] = False,
    temperature: Annotated[
        float,
        typer.Option("--temperature", "-t", help="Sampling temperature.", min=0, max=2),
    ] = 0.7,
    collection: Annotated[
        str,
        typer.Option(
            "--collection",
            "-c",
            help="If set, will put the translated result into a Chroma DB "
            "collection with the name provided.",
        ),
    ] = None,
    diagram_type: Annotated[
        str,
        typer.Option(
            "--diagram-type", "-dg", help="Diagram type to generate in PLANTUML"
        ),
    ] = "Activity",
    add_documentation: Annotated[
        bool,
        typer.Option(
            "--add-documentation/--no-documentation",
            "-ad",
            help="Whether to use documentation in generation",
        ),
    ] = False,
    splitter_type: Annotated[
        str,
        typer.Option(
            "-S",
            "--splitter",
            help="Name of custom splitter to use",
            click_type=click.Choice(list(CUSTOM_SPLITTERS.keys())),
        ),
    ] = "file",
    refiner_types: Annotated[
        list[str],
        typer.Option(
            "-r",
            "--refiner",
            help="List of refiner types to use. Add -r for each refiner to use in\
                refinement chain",
            click_type=click.Choice(list(REFINERS.keys())),
        ),
    ] = ["JanusRefiner"],
    retriever_type: Annotated[
        str,
        typer.Option(
            "-R",
            "--retriever",
            help="Name of custom retriever to use",
            click_type=click.Choice(["active_usings", "language_docs"]),
        ),
    ] = None,
):
    from janus.cli.constants import db_loc, get_collections_config
    from janus.converter.diagram import DiagramGenerator

    refiner_types = [REFINERS[r] for r in refiner_types]
    model_arguments = dict(temperature=temperature)
    collections_config = get_collections_config()
    diagram_generator = DiagramGenerator(
        model=llm_name,
        model_arguments=model_arguments,
        source_language=language,
        max_prompts=max_prompts,
        db_path=db_loc,
        db_config=collections_config,
        splitter_type=splitter_type,
        refiner_types=refiner_types,
        retriever_type=retriever_type,
        diagram_type=diagram_type,
        add_documentation=add_documentation,
    )
    diagram_generator.translate(input_dir, output_dir, failure_dir, overwrite, collection)


def render(
    input_dir: Annotated[
        str,
        typer.Option(
            "--input",
            "-i",
        ),
    ],
    output_dir: Annotated[str, typer.Option("--output", "-o")],
):
    import json
    import subprocess  # nosec

    from janus.cli.constants import homedir

    input_dir = Path(input_dir)
    output_dir = Path(output_dir)
    for input_file in input_dir.rglob("*.json"):
        with open(input_file, "r") as f:
            data = json.load(f)

        output_file = output_dir / input_file.relative_to(input_dir).with_suffix(".txt")
        if not output_file.parent.exists():
            output_file.parent.mkdir()

        text = data["output"].replace("\\n", "\n").strip()
        output_file.write_text(text)

        jar_path = homedir / ".janus/lib/plantuml.jar"
        subprocess.run(["java", "-jar", jar_path, output_file])  # nosec
        output_file.unlink()



from langchain_core.runnables import Runnable, RunnableParallel

from janus.converter.document import Documenter
from janus.parsers.uml import UMLSyntaxParser
from janus.utils.logger import create_logger

log = create_logger(__name__)


class DiagramGenerator(Documenter):
    """A Converter that translates code into a set of PLANTUML diagrams."""

    def __init__(
        self,
        diagram_type="Activity",
        add_documentation=False,
        **kwargs,
    ) -> None:
        """Initialize the DiagramGenerator class

        Arguments:
            diagram_type: type of PLANTUML diagram to generate
            add_documentation: Whether to add a documentation step prior to
                diagram generation.
        """
        self._diagram_type = diagram_type
        self._add_documentation = add_documentation
        self._documenter = Documenter(**kwargs)

        super().__init__(**kwargs)

        self.set_prompt("diagram_with_documentation" if add_documentation else "diagram")
        self._parser = UMLSyntaxParser(language="plantuml")

        self._load_parameters()

    def _load_prompt(self):
        super()._load_prompt()
        self._prompt = self._prompt.partial(DIAGRAM_TYPE=self._diagram_type)

    def _input_runnable(self) -> Runnable:
        if self._add_documentation:
            return RunnableParallel(
                SOURCE_CODE=self._parser.parse_input,
                DOCUMENTATION=self._documenter.chain,
                context=self._retriever,
            )
        return RunnableParallel(
            SOURCE_CODE=self._parser.parse_input,
            context=self._retriever,
        )



import json
import os
from pathlib import Path
from typing import Callable, Protocol, TypeVar

from dotenv import load_dotenv
from langchain_community.llms import HuggingFaceTextGenInference
from langchain_core.runnables import Runnable
from langchain_openai import AzureChatOpenAI, ChatOpenAI

from janus.llm.model_callbacks import (
    COST_PER_1K_TOKENS,
    azure_model_reroutes,
    openai_model_reroutes,
)
from janus.prompts.prompt import (
    ChatGptPromptEngine,
    ClaudePromptEngine,
    CoherePromptEngine,
    Llama2PromptEngine,
    Llama3PromptEngine,
    MistralPromptEngine,
    PromptEngine,
    TitanPromptEngine,
)
from janus.utils.logger import create_logger

log = create_logger(__name__)

try:
    from langchain_community.chat_models import BedrockChat
    from langchain_community.llms.bedrock import Bedrock
except ImportError:
    log.warning(
        "Could not import LangChain's Bedrock Client. If you would like to use Bedrock "
        "models, please install LangChain's Bedrock Client by running 'pip install "
        "janus-llm[bedrock]' or poetry install -E bedrock."
    )

try:
    from langchain_community.llms.huggingface_pipeline import HuggingFacePipeline
except ImportError:
    log.warning(
        "Could not import LangChain's HuggingFace Pipeline Client. If you would like to "
        "use HuggingFace models, please install LangChain's HuggingFace Pipeline Client "
        "by running 'pip install janus-llm[hf-local]' or poetry install -E hf-local."
    )


ModelType = TypeVar(
    "ModelType",
    AzureChatOpenAI,
    ChatOpenAI,
    HuggingFaceTextGenInference,
    Bedrock,
    BedrockChat,
    HuggingFacePipeline,
)


class JanusModelProtocol(Protocol):
    model_id: str
    model_type_name: str
    token_limit: int
    input_token_cost: float
    output_token_cost: float
    prompt_engine: type[PromptEngine]

    def get_num_tokens(self, text: str) -> int:
        ...


class JanusModel(Runnable, JanusModelProtocol):
    ...


load_dotenv()

openai_models = [
    "gpt-4o",
    "gpt-4o-mini",
    "gpt-4",
    "gpt-4-turbo",
    "gpt-4-turbo-preview",
    "gpt-3.5-turbo",
    "gpt-3.5-turbo-16k",
]
azure_models = [
    "gpt-4o",
    "gpt-4o-mini",
    "gpt-3.5-turbo-16k",
]
claude_models = [
    "bedrock-claude-v2",
    "bedrock-claude-instant-v1",
    "bedrock-claude-haiku",
    "bedrock-claude-sonnet",
    "bedrock-claude-sonnet-3.5",
]
llama2_models = [
    "bedrock-llama2-70b",
    "bedrock-llama2-70b-chat",
    "bedrock-llama2-13b",
    "bedrock-llama2-13b-chat",
]
llama3_models = [
    "bedrock-llama3-8b-instruct",
    "bedrock-llama3-70b-instruct",
]
titan_models = [
    "bedrock-titan-text-lite",
    "bedrock-titan-text-express",
    "bedrock-jurassic-2-mid",
    "bedrock-jurassic-2-ultra",
]
cohere_models = [
    "bedrock-command-r-plus",
]
mistral_models = [
    "bedrock-mistral-7b-instruct",
    "bedrock-mistral-large",
    "bedrock-mixtral",
]
bedrock_models = [
    *claude_models,
    *llama2_models,
    *llama3_models,
    *titan_models,
    *cohere_models,
    *mistral_models,
]
all_models = [*azure_models, *bedrock_models]

MODEL_TYPE_CONSTRUCTORS: dict[str, ModelType] = {
    "OpenAI": ChatOpenAI,
    "HuggingFace": HuggingFaceTextGenInference,
    "Azure": AzureChatOpenAI,
    "Bedrock": Bedrock,
    "BedrockChat": BedrockChat,
    "HuggingFaceLocal": HuggingFacePipeline,
}


MODEL_PROMPT_ENGINES: dict[str, Callable[..., PromptEngine]] = {
    **{m: ChatGptPromptEngine for m in openai_models},
    **{m: ChatGptPromptEngine for m in azure_models},
    **{m: ClaudePromptEngine for m in claude_models},
    **{m: Llama2PromptEngine for m in llama2_models},
    **{m: Llama3PromptEngine for m in llama3_models},
    **{m: TitanPromptEngine for m in titan_models},
    **{m: CoherePromptEngine for m in cohere_models},
    **{m: MistralPromptEngine for m in mistral_models},
}

MODEL_ID_TO_LONG_ID = {
    **{m: mr for m, mr in openai_model_reroutes.items()},
    **{m: mr for m, mr in azure_model_reroutes.items()},
    "bedrock-claude-v2": "anthropic.claude-v2",
    "bedrock-claude-instant-v1": "anthropic.claude-instant-v1",
    "bedrock-claude-haiku": "anthropic.claude-3-haiku-20240307-v1:0",
    "bedrock-claude-sonnet": "anthropic.claude-3-sonnet-20240229-v1:0",
    "bedrock-claude-sonnet-3.5": "anthropic.claude-3-5-sonnet-20240620-v1:0",
    "bedrock-llama2-70b": "meta.llama2-70b-v1",
    "bedrock-llama2-70b-chat": "meta.llama2-70b-chat-v1",
    "bedrock-llama2-13b": "meta.llama2-13b-chat-v1",
    "bedrock-llama2-13b-chat": "meta.llama2-13b-v1",
    "bedrock-llama3-8b-instruct": "meta.llama3-8b-instruct-v1:0",
    "bedrock-llama3-70b-instruct": "meta.llama3-70b-instruct-v1:0",
    "bedrock-titan-text-lite": "amazon.titan-text-lite-v1",
    "bedrock-titan-text-express": "amazon.titan-text-express-v1",
    "bedrock-jurassic-2-mid": "ai21.j2-mid-v1",
    "bedrock-jurassic-2-ultra": "ai21.j2-ultra-v1",
    "bedrock-command-r-plus": "cohere.command-r-plus-v1:0",
    "bedrock-mixtral": "mistral.mixtral-8x7b-instruct-v0:1",
    "bedrock-mistral-7b-instruct": "mistral.mistral-7b-instruct-v0:2",
    "bedrock-mistral-large": "mistral.mistral-large-2402-v1:0",
}

MODEL_DEFAULT_ARGUMENTS: dict[str, dict[str, str]] = {
    k: (dict(model_name=k) if k in openai_models else dict(model_id=v))
    for k, v in MODEL_ID_TO_LONG_ID.items()
}

DEFAULT_MODELS = list(MODEL_DEFAULT_ARGUMENTS.keys())

MODEL_CONFIG_DIR = Path.home().expanduser() / ".janus" / "llm"

MODEL_TYPES: dict[str, PromptEngine] = {
    **{m: "OpenAI" for m in openai_models},
    **{m: "Azure" for m in azure_models},
    **{m: "BedrockChat" for m in bedrock_models},
}

TOKEN_LIMITS: dict[str, int] = {
    "gpt-4-32k": 32_768,
    "gpt-4-0613": 8192,
    "gpt-4-1106-preview": 128_000,
    "gpt-4-0125-preview": 128_000,
    "gpt-4o-2024-05-13": 128_000,
    "gpt-4o-2024-08-06": 128_000,
    "gpt-4o-mini": 128_000,
    "gpt-3.5-turbo-0125": 16_384,
    "gpt35-turbo-16k": 16_384,
    "text-embedding-ada-002": 8191,
    "gpt4all": 16_384,
    "anthropic.claude-v2": 100_000,
    "anthropic.claude-instant-v1": 100_000,
    "anthropic.claude-3-haiku-20240307-v1:0": 248_000,
    "anthropic.claude-3-sonnet-20240229-v1:0": 248_000,
    "anthropic.claude-3-5-sonnet-20240620-v1:0": 200_000,
    "meta.llama2-70b-v1": 4096,
    "meta.llama2-70b-chat-v1": 4096,
    "meta.llama2-13b-chat-v1": 4096,
    "meta.llama2-13b-v1": 4096,
    "meta.llama3-8b-instruct-v1:0": 8000,
    "meta.llama3-70b-instruct-v1:0": 8000,
    "amazon.titan-text-lite-v1": 4096,
    "amazon.titan-text-express-v1": 8192,
    "ai21.j2-mid-v1": 8192,
    "ai21.j2-ultra-v1": 8192,
    "cohere.command-r-plus-v1:0": 128_000,
    "mistral.mixtral-8x7b-instruct-v0:1": 32_000,
    "mistral.mistral-7b-instruct-v0:2": 32_000,
    "mistral.mistral-large-2402-v1:0": 32_000,
}


def get_available_model_names() -> list[str]:
    avaialable_models = []
    for file in MODEL_CONFIG_DIR.iterdir():
        if file.is_file():
            avaialable_models.append(MODEL_CONFIG_DIR.stem)
    return avaialable_models


def load_model(model_id) -> JanusModel:
    if not MODEL_CONFIG_DIR.exists():
        MODEL_CONFIG_DIR.mkdir(parents=True)
    model_config_file = MODEL_CONFIG_DIR / f"{model_id}.json"

    if model_config_file.exists():
        log.info(f"Loading {model_id} from {model_config_file}.")
        with open(model_config_file, "r") as f:
            model_config = json.load(f)
        model_type_name = model_config["model_type"]
        model_id = model_config["model_id"]
        model_args = model_config["model_args"]
        token_limit = model_config["token_limit"]
        input_token_cost = model_config["model_cost"]["input"]
        output_token_cost = model_config["model_cost"]["output"]
        input_token_proportion = model_config["input_token_proportion"]

    elif model_id in DEFAULT_MODELS:
        model_id = model_id
        model_long_id = MODEL_ID_TO_LONG_ID[model_id]
        model_type_name = MODEL_TYPES[model_id]
        model_args = MODEL_DEFAULT_ARGUMENTS[model_id]

        token_limit = 0
        input_token_cost = 0.0
        output_token_cost = 0.0
        input_token_proportion = 0.4
        if model_long_id in TOKEN_LIMITS:
            token_limit = TOKEN_LIMITS[model_long_id]
        if model_long_id in COST_PER_1K_TOKENS:
            token_limits = COST_PER_1K_TOKENS[model_long_id]
            input_token_cost = token_limits["input"]
            output_token_cost = token_limits["output"]

    else:
        model_list = "\n\t".join(DEFAULT_MODELS)
        message = (
            f"Model {model_id} not found in user-defined model directory "
            f"({MODEL_CONFIG_DIR}), and is not a default model. Valid default "
            f"models:\n\t{model_list}\n"
            f"To use a custom model, first run `janus llm add`."
        )
        log.error(message)
        raise ValueError(message)

    if model_type_name == "HuggingFaceLocal":
        model = HuggingFacePipeline.from_model_id(
            model_id=model_id,
            task="text-generation",
            model_kwargs=model_args,
        )
        model_args.update(pipeline=model.pipeline)

    elif model_type_name == "OpenAI":
        model_args.update(
            openai_api_key=str(os.getenv("OPENAI_API_KEY")),
        )
        # log.warning("Do NOT use this model in sensitive environments!")
        # log.warning("If you would like to cancel, please press Ctrl+C.")
        # log.warning("Waiting 10 seconds...")
        # Give enough time for the user to read the warnings and cancel
        # time.sleep(10)
        # raise DeprecationWarning("OpenAI models are no longer supported.")

    elif model_type_name == "Azure":
        model_args.update(
            api_key=os.getenv("AZURE_OPENAI_API_KEY"),
            azure_endpoint=os.getenv("AZURE_OPENAI_ENDPOINT"),
            api_version=os.getenv("OPENAI_API_VERSION", "2024-02-01"),
            azure_deployment=model_id,
            request_timeout=3600,
            max_tokens=4096,
        )

    model_type = MODEL_TYPE_CONSTRUCTORS[model_type_name]
    prompt_engine = MODEL_PROMPT_ENGINES[model_id]

    class JanusModel(model_type):
        model_id: str
        # model_name is for LangChain compatibility
        # It searches for `self.model_name` when counting tokens
        model_name: str
        short_model_id: str
        model_type_name: str
        token_limit: int
        input_token_proportion: float
        input_token_cost: float
        output_token_cost: float
        prompt_engine: type[PromptEngine]

    model_args.update(
        model_id=MODEL_ID_TO_LONG_ID[model_id],
        model_name=model_id,  # This is for LangChain compatibility
        short_model_id=model_id,
    )

    return JanusModel(
        model_type_name=model_type_name,
        token_limit=token_limit,
        input_token_cost=input_token_cost,
        output_token_cost=output_token_cost,
        input_token_proportion=input_token_proportion,
        prompt_engine=prompt_engine,
        **model_args,
    )



import logging

import typer
from rich import print

from janus.cli.aggregate import aggregate
from janus.cli.database import db
from janus.cli.diagram import diagram, render
from janus.cli.document import document
from janus.cli.embedding import embedding
from janus.cli.llm import llm
from janus.cli.partition import partition
from janus.cli.self_eval import llm_self_eval
from janus.cli.translate import translate
from janus.metrics.cli import evaluate
from janus.utils.logger import create_logger

httpx_logger = logging.getLogger("httpx")
httpx_logger.setLevel(logging.WARNING)

log = create_logger(__name__)


app = typer.Typer(
    help=(
        "[bold][dark_orange]Janus[/dark_orange] is a CLI for translating, "
        "documenting, and diagramming code using large language models.[/bold]"
    ),
    add_completion=False,
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    rich_markup_mode="rich",
)


def version_callback(value: bool) -> None:
    if value:
        from janus import __version__ as version

        print(f"Janus CLI [blue]v{version}[/blue]")
        raise typer.Exit()


@app.callback()
def common(
    ctx: typer.Context,
    version: bool = typer.Option(
        None,
        "--version",
        "-v",
        callback=version_callback,
        help="Print the version and exit.",
    ),
) -> None:
    """A function for getting the app version

    This will call the version_callback function to print the version and exit.

    Arguments:
        ctx: The typer context
        version: A boolean flag for the version
    """
    pass


aggregate = app.command(
    help=(
        "Aggregate intermediate representations together up to higher levels of "
        "abstraction."
    ),
    no_args_is_help=True,
)(aggregate)

diagram = app.command(
    help="Diagram input code using an LLM.",
    no_args_is_help=True,
)(diagram)

document = app.command(
    help="Document input code using an LLM.",
    no_args_is_help=True,
)(document)

llm_self_eval = app.command(
    help="Use an LLM to evaluate its own performance.",
    no_args_is_help=True,
)(llm_self_eval)

partition = app.command(
    help="Partition input code using an LLM.",
    no_args_is_help=True,
)(partition)

render = app.command(
    help="Render PlantUML from JSON output.",
    no_args_is_help=True,
)(render)

translate = app.command(
    help="Translate code from one language to another using an LLM.",
    no_args_is_help=True,
)(translate)

app.add_typer(db, name="db")
app.add_typer(llm, name="llm")
app.add_typer(evaluate, name="evaluate")
app.add_typer(embedding, name="embedding")


if __name__ == "__main__":
    app()



import functools
import json
import time
from pathlib import Path
from typing import Any

from langchain_core.exceptions import OutputParserException
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.runnables import (
    Runnable,
    RunnableLambda,
    RunnableParallel,
    RunnablePassthrough,
)
from openai import BadRequestError, RateLimitError
from pydantic import ValidationError

from janus.embedding.vectorize import ChromaDBVectorizer
from janus.language.block import CodeBlock, TranslatedCodeBlock
from janus.language.combine import Combiner
from janus.language.naive.registry import CUSTOM_SPLITTERS
from janus.language.splitter import (
    EmptyTreeError,
    FileSizeError,
    Splitter,
    TokenLimitError,
)
from janus.llm.model_callbacks import get_model_callback
from janus.llm.models_info import MODEL_PROMPT_ENGINES, JanusModel, load_model
from janus.parsers.parser import GenericParser, JanusParser, JanusParserException
from janus.refiners.refiner import JanusRefiner

# from janus.refiners.refiner import BasicRefiner, Refiner
from janus.retrievers.retriever import (
    ActiveUsingsRetriever,
    JanusRetriever,
    LanguageDocsRetriever,
)
from janus.utils.enums import LANGUAGES
from janus.utils.logger import create_logger

log = create_logger(__name__)


def run_if_changed(*tracked_vars):
    """Wrapper to skip function calls if the given instance attributes haven't
    been updated. Requires the _changed_attrs set to exist, and the __setattr__
    method to be overridden to track parameter updates in _changed_attrs.
    """

    def wrapper(func):
        @functools.wraps(func)
        def wrapped(self, *args, **kwargs):
            # If there is overlap between the tracked variables and the changed
            #  ones, then call the function as normal
            if not tracked_vars or self._changed_attrs.intersection(tracked_vars):
                func(self, *args, **kwargs)

        return wrapped

    return wrapper


class Converter:
    """Parent class that converts code into something else.

    Children will determine what the code gets converted into. Whether that's translated
    into another language, into pseudocode, requirements, documentation, etc., or
    converted into embeddings
    """

    def __init__(
        self,
        model: str = "gpt-4o",
        model_arguments: dict[str, Any] = {},
        source_language: str = "fortran",
        max_prompts: int = 10,
        max_tokens: int | None = None,
        prompt_template: str = "simple",
        db_path: str | None = None,
        db_config: dict[str, Any] | None = None,
        protected_node_types: tuple[str, ...] = (),
        prune_node_types: tuple[str, ...] = (),
        splitter_type: str = "file",
        refiner_types: list[type[JanusRefiner]] = [JanusRefiner],
        retriever_type: str | None = None,
    ) -> None:
        """Initialize a Converter instance.

        Arguments:
            source_language: The source programming language.
            parser_type: The type of parser to use for parsing the LLM output. Valid
                values are `"code"`, `"text"`, `"eval"`, and `None` (default). If `None`,
                the `Converter` assumes you won't be parsing an output (i.e., adding to an
                embedding DB).
            max_prompts: The maximum number of prompts to try before giving up.
            max_tokens: The maximum number of tokens to use in the LLM. If `None`, the
                converter will use half the model's token limit.
            prompt_template: The name of the prompt template to use.
            db_path: The path to the database to use for vectorization.
            db_config: The configuration for the database.
            protected_node_types: A set of node types that aren't to be merged.
            prune_node_types: A set of node types which should be pruned.
            splitter_type: The type of splitter to use. Valid values are `"file"`,
                `"tag"`, `"chunk"`, `"ast-strict"`, and `"ast-flex"`.
            refiner_type: The type of refiner to use. Valid values:
                - "parser"
                - "reflection"
                - None
            retriever_type: The type of retriever to use. Valid values:
                - "active_usings"
                - "language_docs"
                - None
        """
        self._changed_attrs: set = set()

        self.max_prompts: int = max_prompts
        self._max_tokens: int | None = max_tokens
        self.override_token_limit: bool = max_tokens is not None

        self._model_name: str
        self._custom_model_arguments: dict[str, Any]

        self._source_language: str
        self._source_suffixes: list[str]

        self._target_language = "json"
        self._target_suffix = ".json"

        self._protected_node_types: tuple[str, ...] = ()
        self._prune_node_types: tuple[str, ...] = ()
        self._max_tokens: int | None = max_tokens
        self._prompt_template_name: str
        self._db_path: str | None
        self._db_config: dict[str, Any] | None

        self._llm: JanusModel
        self._prompt: ChatPromptTemplate

        self._parser: JanusParser = GenericParser()
        self._base_parser: JanusParser = GenericParser()
        self._combiner: Combiner = Combiner()

        self._splitter_type: str
        self._refiner_types: list[type[JanusRefiner]]
        self._retriever_type: str | None

        self._splitter: Splitter
        self._refiner: JanusRefiner
        self._retriever: JanusRetriever

        self.set_splitter(splitter_type=splitter_type)
        self.set_refiner_types(refiner_types=refiner_types)
        self.set_retriever(retriever_type=retriever_type)
        self.set_model(model_name=model, **model_arguments)
        self.set_prompt(prompt_template=prompt_template)
        self.set_source_language(source_language)
        self.set_protected_node_types(protected_node_types)
        self.set_prune_node_types(prune_node_types)
        self.set_db_path(db_path=db_path)
        self.set_db_config(db_config=db_config)

        # Child class must call this. Should we enforce somehow?
        # self._load_parameters()

    def __setattr__(self, key: Any, value: Any) -> None:
        if hasattr(self, "_changed_attrs"):
            if not hasattr(self, key) or getattr(self, key) != value:
                self._changed_attrs.add(key)
        # Avoid infinite recursion
        elif key != "_changed_attrs":
            self._changed_attrs = set()
        super().__setattr__(key, value)

    def _load_parameters(self) -> None:
        self._load_model()
        self._load_prompt()
        self._load_retriever()
        self._load_refiner_chain()
        self._load_splitter()
        self._load_vectorizer()
        self._load_chain()
        self._changed_attrs.clear()

    def set_model(self, model_name: str, **custom_arguments: dict[str, Any]):
        """Validate and set the model name.

        The affected objects will not be updated until translate() is called.

        Arguments:
            model_name: The name of the model to use. Valid models are found in
                `janus.llm.models_info.MODEL_CONSTRUCTORS`.
            custom_arguments: Additional arguments to pass to the model constructor.
        """
        self._model_name = model_name
        self._custom_model_arguments = custom_arguments

    def set_prompt(self, prompt_template: str) -> None:
        """Validate and set the prompt template name.

        Arguments:
            prompt_template: name of prompt template directory
                (see janus/prompts/templates) or path to a directory.
        """
        self._prompt_template_name = prompt_template

    def set_splitter(self, splitter_type: str) -> None:
        """Validate and set the prompt template name.

        Arguments:
            prompt_template: name of prompt template directory
                (see janus/prompts/templates) or path to a directory.
        """
        if splitter_type not in CUSTOM_SPLITTERS:
            raise ValueError(f'Splitter type "{splitter_type}" does not exist.')

        self._splitter_type = splitter_type

    def set_refiner_types(self, refiner_types: list[type[JanusRefiner]]) -> None:
        """Validate and set the refiner type

        Arguments:
            refiner_type: the type of refiner to use
        """
        self._refiner_types = refiner_types

    def set_retriever(self, retriever_type: str | None) -> None:
        """Validate and set the retriever type

        Arguments:
            retriever_type: the type of retriever to use
        """
        self._retriever_type = retriever_type

    def set_source_language(self, source_language: str) -> None:
        """Validate and set the source language.

        Arguments:
            source_language: The source programming language.
        """
        source_language = source_language.lower()
        if source_language not in LANGUAGES:
            raise ValueError(
                f"Invalid source language: {source_language}. "
                "Valid source languages are found in `janus.utils.enums.LANGUAGES`."
            )

        self._source_suffixes = [
            f".{ext}" for ext in LANGUAGES[source_language]["suffixes"]
        ]

        self._source_language = source_language

    def set_protected_node_types(self, protected_node_types: tuple[str, ...]) -> None:
        """Set the protected (non-mergeable) node types. This will often be structures
        like functions, classes, or modules which you might want to keep separate

        The affected objects will not be updated until _load_parameters() is called.

        Arguments:
            protected_node_types: A set of node types that aren't to be merged
        """
        self._protected_node_types = tuple(set(protected_node_types or []))

    def set_prune_node_types(self, prune_node_types: tuple[str, ...]) -> None:
        """Set the node types to prune. This will often be structures
        like comments or whitespace which you might want to keep out of the LLM

        The affected objects will not be updated until _load_parameters() is called.

        Arguments:
            prune_node_types: A set of node types which should be pruned
        """
        self._prune_node_types = tuple(set(prune_node_types or []))

    def set_db_path(self, db_path: str | None) -> None:
        self._db_path = db_path

    def set_db_config(self, db_config: dict[str, Any] | None) -> None:
        self._db_config = db_config

    @run_if_changed(
        "_source_language",
        "_max_tokens",
        "_llm",
        "_protected_node_types",
        "_prune_node_types",
        "_custom_splitter",
    )
    def _load_splitter(self) -> None:
        """Load the splitter according to this instance's attributes.

        If the relevant fields have not been changed since the last time this method was
        called, nothing happens.
        """
        kwargs: dict[str, Any] = dict(
            language=self._source_language,
            max_tokens=self._max_tokens,
            model=self._llm,
            protected_node_types=self._protected_node_types,
            prune_node_types=self._prune_node_types,
        )

        if self._splitter_type == "tag":
            kwargs["tag"] = "<ITMOD_ALC_SPLIT>"  # Hardcoded for now

        self._splitter = CUSTOM_SPLITTERS[self._splitter_type](**kwargs)

    @run_if_changed("_model_name", "_custom_model_arguments")
    def _load_model(self) -> None:
        """Load the model according to this instance's attributes.

        If the relevant fields have not been changed since the last time this method was
        called, nothing happens.
        """

        # Get default arguments, set custom ones
        # model_arguments = deepcopy(MODEL_DEFAULT_ARGUMENTS[self._model_name])
        # model_arguments.update(self._custom_model_arguments)

        # Load the model
        self._llm = load_model(self._model_name)
        token_limit = self._llm.token_limit

        # Set the max_tokens to less than half the model's limit to allow for enough
        # tokens at output
        # Only modify max_tokens if it is not specified by user
        if not self.override_token_limit:
            self._max_tokens = int(token_limit * self._llm.input_token_proportion)

    @run_if_changed(
        "_prompt_template_name",
        "_source_language",
        "_model_name",
        "_parser",
    )
    def _load_prompt(self) -> None:
        """Load the prompt according to this instance's attributes.

        If the relevant fields have not been changed since the last time this
        method was called, nothing happens.
        """
        prompt_engine = MODEL_PROMPT_ENGINES[self._llm.short_model_id](
            source_language=self._source_language,
            prompt_template=self._prompt_template_name,
        )
        self._prompt = prompt_engine.prompt
        self._prompt = self._prompt.partial(
            format_instructions=self._parser.get_format_instructions()
        )

    @run_if_changed("_db_path", "_db_config")
    def _load_vectorizer(self) -> None:
        if self._db_path is None or self._db_config is None:
            self._vectorizer = None
            return
        vectorizer_factory = ChromaDBVectorizer()
        self._vectorizer = vectorizer_factory.create_vectorizer(
            self._db_path, self._db_config
        )

    @run_if_changed("_retriever_type")
    def _load_retriever(self):
        if self._retriever_type == "active_usings":
            self._retriever = ActiveUsingsRetriever()
        elif self._retriever_type == "language_docs":
            self._retriever = LanguageDocsRetriever(self._llm, self._source_language)
        else:
            self._retriever = JanusRetriever()

    @run_if_changed("_refiner_types", "_model_name", "max_prompts", "_parser")
    def _load_refiner_chain(self) -> None:
        self._refiner_chain = RunnableParallel(
            completion=self._llm,
            prompt_value=RunnablePassthrough(),
        )
        for refiner_type in self._refiner_types[:-1]:
            # NOTE: Do NOT remove refiner_type=refiner_type from lambda.
            # Due to lambda capture, must be present or chain will not
            # be correctly constructed.
            self._refiner_chain = self._refiner_chain | RunnableParallel(
                completion=lambda x, refiner_type=refiner_type: refiner_type(
                    llm=self._llm,
                    parser=self._base_parser,
                    max_retries=self.max_prompts,
                ).parse_completion(**x),
                prompt_value=lambda x: x["prompt_value"],
            )
        self._refiner_chain = self._refiner_chain | RunnableLambda(
            lambda x: self._refiner_types[-1](
                llm=self._llm,
                parser=self._parser,
                max_retries=self.max_prompts,
            ).parse_completion(**x)
        )

    @run_if_changed("_parser", "_retriever", "_prompt", "_llm", "_refiner_chain")
    def _load_chain(self):
        self.chain = self._input_runnable() | self._prompt | self._refiner_chain

    def _input_runnable(self) -> Runnable:
        return RunnableParallel(
            SOURCE_CODE=self._parser.parse_input,
            context=self._retriever,
        )

    def translate(
        self,
        input_directory: str | Path,
        output_directory: str | Path | None = None,
        failure_directory: str | Path | None = None,
        overwrite: bool = False,
        collection_name: str | None = None,
    ) -> None:
        """Convert code in the input directory from the source language to the target
        language, and write the resulting files to the output directory.

        Arguments:
            input_directory: The directory containing the code to translate.
            output_directory: The directory to write the translated code to.
            overwrite: Whether to overwrite existing files (vs skip them)
            collection_name: Collection to add to
        """
        # Convert paths to pathlib Paths if needed
        if isinstance(input_directory, str):
            input_directory = Path(input_directory)
        if isinstance(output_directory, str):
            output_directory = Path(output_directory)
        if isinstance(failure_directory, str):
            failure_directory = Path(failure_directory)

        # Make sure the output directory exists
        if output_directory is not None and not output_directory.exists():
            output_directory.mkdir(parents=True)
        if failure_directory is not None and not failure_directory.exists():
            failure_directory.mkdir(parents=True)

        input_paths = []
        for ext in self._source_suffixes:
            input_paths.extend(input_directory.rglob(f"**/*{ext}"))

        log.info(f"Input directory: {input_directory.absolute()}")
        log.info(
            f"{self._source_language} {self._source_suffixes} files: "
            f"{len(input_paths)}"
        )
        log.info(
            "Other files (skipped): "
            f"{len(list(input_directory.iterdir())) - len(input_paths)}\n"
        )
        if output_directory is not None:
            output_paths = [
                output_directory
                / p.relative_to(input_directory).with_suffix(self._target_suffix)
                for p in input_paths
            ]
        else:
            output_paths = [None for _ in input_paths]

        if failure_directory is not None:
            failure_paths = [
                failure_directory
                / p.relative_to(input_directory).with_suffix(self._target_suffix)
                for p in input_paths
            ]
        else:
            failure_paths = [None for _ in input_paths]
        in_out_pairs = list(zip(input_paths, output_paths, failure_paths))
        if not overwrite:
            n_files = len(in_out_pairs)
            in_out_pairs = [
                (inp, outp, failp)
                for inp, outp, failp in in_out_pairs
                if outp is None or not outp.exists()
            ]
            log.info(
                f"Skipping {n_files - len(in_out_pairs)} existing "
                f"{self._source_suffixes} files"
            )
        log.info(f"Translating {len(in_out_pairs)} {self._source_suffixes} files")

        # Loop through each input file, convert and save it
        total_cost = 0.0
        for in_path, out_path, fail_path in in_out_pairs:
            # Translate the file, skip it if there's a rate limit error
            log.info(f"Processing {in_path.relative_to(input_directory)}")
            out_block = self.translate_file(in_path, fail_path)
            total_cost += out_block.total_cost
            log.info(f"Current Running Cost: {total_cost}")

            # Don't attempt to write files for which translation failed
            if not out_block.translated:
                continue

            if collection_name is not None:
                self._vectorizer.add_nodes_recursively(
                    out_block,
                    collection_name,
                    in_path.name,
                )

            # Make sure the tree's code has been consolidated at the top level
            #  before writing to file
            self._combiner.combine(out_block)
            if out_path is not None and (overwrite or not out_path.exists()):
                self._save_to_file(out_block, out_path)

        log.info(f"Total cost: ${total_cost:,.2f}")

    def translate_file(
        self, file: Path, failure_path: Path | None = None
    ) -> TranslatedCodeBlock:
        """Translate a single file.

        Arguments:
            file: Input path to file
            failure_path: path to directory to store failure summaries`

        Returns:
            A `TranslatedCodeBlock` object. This block does not have a path set, and its
            code is not guaranteed to be consolidated. To amend this, run
            `Combiner.combine_children` on the block.
        """
        self._load_parameters()
        filename = file.name

        input_block = self._split_file(file)
        t0 = time.time()
        output_block = self._iterative_translate(input_block, failure_path)
        output_block.processing_time = time.time() - t0
        if output_block.translated:
            completeness = output_block.translation_completeness
            log.info(
                f"[{filename}] Translation complete\n"
                f"  {completeness:.2%} of input successfully translated\n"
                f"  Total cost: ${output_block.total_cost:,.2f}\n"
                f"  Output CodeBlock Structure:\n{input_block.tree_str()}\n"
            )

        else:
            log.error(
                f"[{filename}] Translation failed\n"
                f"  Total cost: ${output_block.total_cost:,.2f}\n"
            )
        return output_block

    def _iterative_translate(
        self, root: CodeBlock, failure_path: Path | None = None
    ) -> TranslatedCodeBlock:
        """Translate the passed CodeBlock representing a full file.

        Arguments:
            root: A root block representing the top-level block of a file
            failure_path: path to store data files for failed translations

        Returns:
            A `TranslatedCodeBlock`
        """
        translated_root = TranslatedCodeBlock(root, self._target_language)
        last_prog, prog_delta = 0, 0.1
        stack = [translated_root]
        try:
            while stack:
                translated_block = stack.pop()

                self._add_translation(translated_block)

                # If translating this block was unsuccessful, don't bother with its
                #  children (they wouldn't show up in the final text anyway)
                if not translated_block.translated:
                    continue

                stack.extend(translated_block.children)

                progress = translated_root.translation_completeness
                if progress - last_prog > prog_delta:
                    last_prog = int(progress / prog_delta) * prog_delta
                    log.info(f"[{root.name}] progress: {progress:.2%}")
        except RateLimitError:
            pass
        except OutputParserException as e:
            log.error(f"Skipping file, failed to parse output: {e}.")
        except BadRequestError as e:
            if str(e).startswith("Detected an error in the prompt"):
                log.warning("Malformed input, skipping")
            raise e
        except ValidationError as e:
            # Only allow ValidationError to pass if token limit is manually set
            if self.override_token_limit:
                log.warning(
                    "Current file and manually set token "
                    "limit is too large for this model, skipping"
                )
            raise e
        except TokenLimitError:
            log.warning("Ran into irreducible node too large for context, skipping")
        except EmptyTreeError:
            log.warning("Input file has no nodes of interest, skipping")
        except FileSizeError:
            log.warning("Current tile is too large for basic splitter, skipping")
        except ValueError as e:
            if str(e).startswith(
                "Error raised by bedrock service"
            ) and "maximum context length" in str(e):
                log.warning(
                    "Input is too large for this model's context length, skipping"
                )
            raise e
        finally:
            log.debug(
                f"Resulting Block: {json.dumps(self._get_output_obj(translated_root))}"
            )
            if not translated_root.translated:
                if failure_path is not None:
                    self._save_to_file(translated_root, failure_path)

        return translated_root

    def _add_translation(self, block: TranslatedCodeBlock) -> None:
        """Given an "empty" `TranslatedCodeBlock`, translate the code represented in
        `block.original`, setting the relevant fields in the translated block. The
        `TranslatedCodeBlock` is updated in-pace, nothing is returned. Note that this
        translates *only* the code for this block, not its children.

        Arguments:
            block: An empty `TranslatedCodeBlock`
        """
        if block.translated:
            return

        if block.original.text is None:
            block.translated = True
            return

        if self._llm is None:
            message = (
                "Model not configured correctly, cannot translate. Try setting "
                "the model"
            )
            log.error(message)
            raise ValueError(message)

        log.debug(f"[{block.name}] Translating...")
        log.debug(f"[{block.name}] Input text:\n{block.original.text}")

        # Track the cost of translating this block
        #  TODO: If non-OpenAI models with prices are added, this will need
        #   to be updated.
        with get_model_callback() as cb:
            try:
                t0 = time.time()
                block.text = self._run_chain(block)
            except JanusParserException as e:
                block.text = e.unparsed_output
                block.tokens = self._llm.get_num_tokens(block.text)
                raise e
            finally:
                block.processing_time = time.time() - t0
                block.cost = cb.total_cost
                block.request_input_tokens = cb.prompt_tokens
                block.request_output_tokens = cb.completion_tokens
                block.num_requests = cb.successful_requests

        block.tokens = self._llm.get_num_tokens(block.text)
        block.translated = True

        log.debug(f"[{block.name}] Output code:\n{block.text}")

    def _split_file(self, file: Path) -> CodeBlock:
        filename = file.name
        log.info(f"[{filename}] Splitting file")
        root = self._splitter.split(file)
        log.info(
            f"[{filename}] File split into {root.n_descendents:,} blocks, "
            f"tree of height {root.height}"
        )
        log.info(f"[{filename}] Input CodeBlock Structure:\n{root.tree_str()}")
        return root

    def _run_chain(self, block: TranslatedCodeBlock) -> str:
        return self.chain.invoke(block.original)

    def _get_output_obj(
        self, block: TranslatedCodeBlock
    ) -> dict[str, int | float | str | dict[str, str] | dict[str, float]]:
        output_obj: str | dict[str, str]
        if not block.translation_completed:
            # translation wasn't completed, so combined parsing will likely fail
            output_obj = block.complete_text
        else:
            output_str = self._parser.parse_combined_output(block.complete_text)
            try:
                output_obj = json.loads(output_str)
            except json.JSONDecodeError:
                output_obj = output_str

        return dict(
            input=block.original.text or "",
            metadata=dict(
                cost=block.total_cost,
                processing_time=block.processing_time,
                num_requests=block.total_num_requests,
                input_tokens=block.total_request_input_tokens,
                output_tokens=block.total_request_output_tokens,
            ),
            output=output_obj,
        )

    def _save_to_file(self, block: TranslatedCodeBlock, out_path: Path) -> None:
        """Save a file to disk.

        Arguments:
            block: The `TranslatedCodeBlock` to save to a file.
        """
        obj = self._get_output_obj(block)
        out_path.parent.mkdir(parents=True, exist_ok=True)
        out_path.write_text(json.dumps(obj, indent=2), encoding="utf-8")



from copy import deepcopy
from typing import List

from janus.converter.converter import Converter
from janus.language.block import CodeBlock, TranslatedCodeBlock


class Aggregator(Converter):
    def __init__(
        self,
        intermediate_converters: List[Converter],
        separator: str = "\n==============\n",
        **kwargs,
    ):
        """
        Initialization Method
        Arguments:
            intermediate_converters - list of converters to use
            separator: separator string to partition different outputs before combination
        """
        self._intermediate_converters = intermediate_converters
        self._separator = separator
        super().__init__(**kwargs)
        self._load_parameters()

    def _iterative_translate(self, root: CodeBlock) -> TranslatedCodeBlock:
        res = TranslatedCodeBlock(root, language=self._target_language)
        return self._recursive_translate(res)

    def _recursive_translate(self, root: TranslatedCodeBlock) -> None:
        """
        Recursively translates code blocks from a bottom up approach
        """
        original_text = root.original.text
        if len(root.children) > 0:
            for c in root.children:
                self._recursive_translate(c)
            root.original.text = self._combine_blocks(root.children, self._separator)
        else:
            int_reps = [deepcopy(root) for ic in self._intermediate_converters]
            for ic, r in zip(self._intermediate_converters, int_reps):
                ic._add_translation(r)
            root.original.text = self._combine_blocks(int_reps, self._separator)
        self._add_translation(root)
        root.original.text = original_text
        return root

    def _combine_blocks(self, blocks: List[TranslatedCodeBlock], separator: str) -> str:
        """
        Combines code blocks into a single piece of text
        """
        return separator.join([block.text for block in blocks])



from pathlib import Path
from typing import List, Optional

import click
import typer
from typing_extensions import Annotated

from janus.cli.constants import get_subclasses
from janus.converter.converter import Converter
from janus.language.naive.registry import CUSTOM_SPLITTERS
from janus.utils.enums import LANGUAGES


def aggregate(
    input_dir: Annotated[
        Path,
        typer.Option(
            "--input",
            "-i",
            help="The directory containing the source code to be translated. "
            "The files should all be in one flat directory.",
        ),
    ],
    language: Annotated[
        str,
        typer.Option(
            "--language",
            "-l",
            help="The language of the source code.",
            click_type=click.Choice(sorted(LANGUAGES)),
        ),
    ],
    output_dir: Annotated[
        Path,
        typer.Option(
            "--output-dir", "-o", help="The directory to store the translated code in."
        ),
    ],
    llm_name: Annotated[
        str,
        typer.Option(
            "--llm",
            "-L",
            help="The custom name of the model set with 'janus llm add'.",
        ),
    ],
    failure_dir: Annotated[
        Optional[Path],
        typer.Option(
            "--failure-directory",
            "-f",
            help="The directory to store failure files during translation",
        ),
    ] = None,
    max_prompts: Annotated[
        int,
        typer.Option(
            "--max-prompts",
            "-m",
            help="The maximum number of times to prompt a model on one functional block "
            "before exiting the application. This is to prevent wasting too much money.",
        ),
    ] = 10,
    overwrite: Annotated[
        bool,
        typer.Option(
            "--overwrite/--preserve",
            help="Whether to overwrite existing files in the output directory",
        ),
    ] = False,
    temperature: Annotated[
        float,
        typer.Option("--temperature", "-t", help="Sampling temperature.", min=0, max=2),
    ] = 0.7,
    collection: Annotated[
        str,
        typer.Option(
            "--collection",
            "-c",
            help="If set, will put the translated result into a Chroma DB "
            "collection with the name provided.",
        ),
    ] = None,
    splitter_type: Annotated[
        str,
        typer.Option(
            "-S",
            "--splitter",
            help="Name of custom splitter to use",
            click_type=click.Choice(list(CUSTOM_SPLITTERS.keys())),
        ),
    ] = "file",
    intermediate_converters: Annotated[
        List[str],
        typer.Option(
            "-C",
            "--converter",
            help="Name of an intermediate converter to use",
            click_type=click.Choice([c.__name__ for c in get_subclasses(Converter)]),
        ),
    ] = ["Documenter"],
):
    from janus.cli.constants import db_loc, get_collections_config
    from janus.converter.aggregator import Aggregator

    converter_subclasses = get_subclasses(Converter)
    converter_subclasses_map = {c.__name__: c for c in converter_subclasses}
    model_arguments = dict(temperature=temperature)
    collections_config = get_collections_config()
    converters = []
    for ic in intermediate_converters:
        converters.append(
            converter_subclasses_map[ic](
                model=llm_name,
                model_arguments=model_arguments,
                source_language=language,
                max_prompts=max_prompts,
                db_path=db_loc,
                db_config=collections_config,
                splitter_type=splitter_type,
            )
        )

    aggregator = Aggregator(
        intermediate_converters=converters,
        model=llm_name,
        model_arguments=model_arguments,
        source_language=language,
        max_prompts=max_prompts,
        db_path=db_loc,
        db_config=collections_config,
        splitter_type=splitter_type,
        prompt_template="basic_aggregation",
    )
    aggregator.translate(input_dir, output_dir, failure_dir, overwrite, collection)



import re

from langchain_core.messages import BaseMessage

from janus.parsers.parser import JanusParser, JanusParserException
from janus.utils.logger import create_logger

log = create_logger(__name__)


class CodeParser(JanusParser):
    language: str

    def parse(self, text: str | BaseMessage) -> str:
        if isinstance(text, BaseMessage):
            text = str(text.content)
        pattern = rf"```[^\S\r\n]*(?:{self.language}[^\S\r\n]*)?\n?(.*?)\n*```"
        code = re.search(pattern, text, re.DOTALL)
        if code is None:
            raise JanusParserException(
                text,
                "Code not find code between triple square brackets",
            )
        return str(code.group(1))

    def get_format_instructions(self) -> str:
        return "Output must contain text contained within triple square brackets (```)"



import json
import re
from typing import Any

from langchain.output_parsers import PydanticOutputParser
from langchain_core.exceptions import OutputParserException
from langchain_core.messages import BaseMessage
from langchain_core.pydantic_v1 import BaseModel, Field, conint

from janus.language.block import CodeBlock
from janus.parsers.parser import JanusParser
from janus.utils.logger import create_logger

log = create_logger(__name__)


class Criteria(BaseModel):
    reasoning: str = Field(description="A short explanation for the given score")
    # Constrained to an integer between 1 and 4
    score: conint(ge=1, le=4) = Field(  # type: ignore
        description="An integer score between 1 and 4 (inclusive), 4 being the best"
    )


class Comment(BaseModel):
    comment_id: str = Field(description="The 8-character comment ID")
    completeness: Criteria = Field(description="The completeness of the comment")
    hallucination: Criteria = Field(description="The factualness of the comment")
    readability: Criteria = Field(description="The readability of the comment")
    usefulness: Criteria = Field(description="The usefulness of the comment")


class CommentList(BaseModel):
    __root__: list[Comment] = Field(
        description=(
            "A list of inline comment evaluations. Each element should include"
            " the comment's 8-character ID in the `comment_id` field, and four"
            " score objects corresponding to each metric (`completeness`,"
            " `hallucination`, `readability`, and `usefulness`)."
        )
    )


class InlineCommentParser(JanusParser, PydanticOutputParser):
    comments: dict[str, str]

    def __init__(self):
        PydanticOutputParser.__init__(
            self,
            pydantic_object=CommentList,
            comments=[],
        )

    def parse_input(self, block: CodeBlock) -> str:
        # TODO: Perform comment stripping/placeholding here rather than in script
        text = super().parse_input(block)
        self.comments = dict(
            re.findall(
                r"<(?:BLOCK|INLINE)_COMMENT (\w{8})> (.*)$",
                text,
                flags=re.MULTILINE,
            )
        )
        return text

    def parse(self, text: str | BaseMessage) -> str:
        if isinstance(text, BaseMessage):
            text = str(text.content)

        # Strip everything outside the JSON object
        begin, end = text.find("["), text.rfind("]")
        text = text[begin : end + 1]

        try:
            out: CommentList = super().parse(text)
        except json.JSONDecodeError as e:
            log.debug(f"Invalid JSON object. Output:\n{text}")
            raise OutputParserException(f"Got invalid JSON object. Error: {e}")

        evals: dict[str, Any] = {c.comment_id: c.dict() for c in out.__root__}

        seen_keys = set(evals.keys())
        expected_keys = set(self.comments.keys())
        missing_keys = expected_keys.difference(seen_keys)
        invalid_keys = seen_keys.difference(expected_keys)
        if missing_keys:
            log.debug(f"Missing keys: {missing_keys}")
            if invalid_keys:
                log.debug(f"Invalid keys: {invalid_keys}")
            log.debug(f"Missing keys: {missing_keys}")
            raise OutputParserException(
                f"Got invalid return object. Missing the following expected "
                f"keys: {missing_keys}"
            )

        for key in invalid_keys:
            del evals[key]

        for cid in evals.keys():
            evals[cid]["comment"] = self.comments[cid]
            evals[cid].pop("comment_id")

        return json.dumps(evals)

    def parse_combined_output(self, text: str) -> str:
        if not text.strip():
            return str({})
        objs = [json.loads(line.strip()) for line in text.split("\n") if line.strip()]
        output_obj = {}
        for obj in objs:
            output_obj.update(obj)
        return json.dumps(output_obj)



import re
from typing import Any

from langchain.output_parsers import RetryWithErrorOutputParser
from langchain_core.exceptions import OutputParserException
from langchain_core.messages import AIMessage
from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompt_values import PromptValue
from langchain_core.runnables import RunnableSerializable

from janus.llm.models_info import MODEL_PROMPT_ENGINES, JanusModel
from janus.parsers.parser import JanusParser
from janus.utils.logger import create_logger

log = create_logger(__name__)


class JanusRefiner(JanusParser):
    parser: JanusParser

    def parse_runnable(self, input: dict[str, Any]) -> Any:
        return self.parse_completion(**input)

    def parse_completion(self, completion: str, **kwargs) -> Any:
        return self.parser.parse(completion)

    def parse(self, text: str) -> str:
        raise NotImplementedError


class SimpleRetry(JanusRefiner):
    max_retries: int
    retry_chain: RunnableSerializable

    def __init__(
        self,
        llm: JanusModel,
        parser: JanusParser,
        max_retries: int,
    ):
        retry_chain = llm | StrOutputParser()
        super().__init__(
            retry_chain=retry_chain,
            parser=parser,
            max_retries=max_retries,
        )

    def parse_completion(
        self, completion: str, prompt_value: PromptValue, **kwargs
    ) -> Any:
        for retry_number in range(self.max_retries):
            try:
                return self.parser.parse(completion)
            except OutputParserException:
                completion = self.retry_chain.invoke(prompt_value)

        return self.parser.parse(completion)


class FixParserExceptions(JanusRefiner, RetryWithErrorOutputParser):
    def __init__(self, llm: JanusModel, parser: JanusParser, max_retries: int):
        retry_prompt = MODEL_PROMPT_ENGINES[llm.short_model_id](
            source_language="text",
            prompt_template="refinement/fix_exceptions",
        ).prompt
        chain = retry_prompt | llm | StrOutputParser()
        RetryWithErrorOutputParser.__init__(
            self, parser=parser, retry_chain=chain, max_retries=max_retries
        )

    def parse_completion(
        self, completion: str, prompt_value: PromptValue, **kwargs
    ) -> Any:
        return self.parse_with_prompt(completion, prompt_value=prompt_value)


class ReflectionRefiner(JanusRefiner):
    max_retries: int
    reflection_chain: RunnableSerializable
    revision_chain: RunnableSerializable
    reflection_prompt_name: str

    def __init__(
        self,
        llm: JanusModel,
        parser: JanusParser,
        max_retries: int,
        prompt_template_name: str = "refinement/reflection",
    ):
        reflection_prompt = MODEL_PROMPT_ENGINES[llm.short_model_id](
            source_language="text",
            prompt_template=prompt_template_name,
        ).prompt
        revision_prompt = MODEL_PROMPT_ENGINES[llm.short_model_id](
            source_language="text",
            prompt_template="refinement/revision",
        ).prompt

        reflection_chain = reflection_prompt | llm | StrOutputParser()
        revision_chain = revision_prompt | llm | StrOutputParser()
        super().__init__(
            reflection_prompt_name=prompt_template_name,
            reflection_chain=reflection_chain,
            revision_chain=revision_chain,
            parser=parser,
            max_retries=max_retries,
        )

    def parse_completion(
        self, completion: str, prompt_value: PromptValue, **kwargs
    ) -> Any:
        log.debug(f"Reflection Prompt: {self.reflection_prompt_name}")
        for retry_number in range(self.max_retries):
            reflection = self.reflection_chain.invoke(
                dict(
                    prompt=prompt_value.to_string(),
                    completion=completion,
                )
            )
            if re.search(r"\bLGTM\b", reflection) is not None:
                return self.parser.parse(completion)
            if not retry_number:
                log.debug(f"Completion:\n{completion}")
            log.debug(f"Reflection:\n{reflection}")
            completion = self.revision_chain.invoke(
                dict(
                    prompt=prompt_value.to_string(),
                    completion=completion,
                    reflection=reflection,
                )
            )
            log.debug(f"Revision:\n{completion}")

        return self.parser.parse(completion)


class RequirementsReflectionRefiner(JanusRefiner):
    """
    This requirements-specific refiner is intended to address a common issue with
    requirements reflection, where over the course of several reflection loops,
    requirements lists grow increasingly verbose, eventually becoming too wordy
    to be useful. To reduce this, this refiner interlaces an additional reflection
    -> revision loop which de-duplicates requirements.
    """

    max_retries: int
    reflection_chain: RunnableSerializable
    revision_chain: RunnableSerializable
    reflect_duplication_chain: RunnableSerializable
    revise_duplication_chain: RunnableSerializable
    reflection_prompt_name: str

    def __init__(
        self,
        llm: JanusModel,
        parser: JanusParser,
        max_retries: int,
        prompt_template_name: str = "refinement/reflection/incose",
    ):
        # Main reflection loop
        reflection_prompt = MODEL_PROMPT_ENGINES[llm.short_model_id](
            source_language="text",
            prompt_template="refinement/reflection/incose",
        ).prompt
        revision_prompt = MODEL_PROMPT_ENGINES[llm.short_model_id](
            source_language="text",
            prompt_template="refinement/revision/incose",
        ).prompt
        # De-duplication loop
        reflect_duplication_prompt = MODEL_PROMPT_ENGINES[llm.short_model_id](
            source_language="text",
            prompt_template="refinement/reflection/incose_deduplicate",
        ).prompt
        revise_duplication_prompt = MODEL_PROMPT_ENGINES[llm.short_model_id](
            source_language="text",
            prompt_template="refinement/revision/incose_deduplicate",
        ).prompt

        reflection_chain = reflection_prompt | llm | StrOutputParser()
        revision_chain = revision_prompt | llm | StrOutputParser()
        reflect_duplication_chain = reflect_duplication_prompt | llm | StrOutputParser()
        revise_duplication_chain = revise_duplication_prompt | llm | StrOutputParser()
        super().__init__(
            reflection_prompt_name=prompt_template_name,
            reflection_chain=reflection_chain,
            revision_chain=revision_chain,
            reflect_duplication_chain=reflect_duplication_chain,
            revise_duplication_chain=revise_duplication_chain,
            parser=parser,
            max_retries=max_retries,
        )

    def parse_completion(
        self, completion: str, prompt_value: PromptValue, **kwargs
    ) -> Any:
        log.debug(f"Reflection Prompt: {self.reflection_prompt_name}")
        if isinstance(completion, AIMessage):
            completion = completion.content
        for retry_number in range(self.max_retries):
            # First, check if the generated requirements are redundant or too specific
            duplication_reflection = self.reflect_duplication_chain.invoke(
                dict(
                    prompt=prompt_value.to_string(),
                    completion=completion,
                )
            )
            if re.search(r"\bLGTM\b", duplication_reflection) is not None:
                log.debug(
                    "No de-duplication suggested in reflection, "
                    "passing to next reflection step"
                )
            else:
                completion = self.revise_duplication_chain.invoke(
                    dict(
                        prompt=prompt_value.to_string(),
                        completion=completion,
                        reflection=duplication_reflection,
                    )
                )

            # Once we're happy with the results or trimmed them down,
            # continue with the typical reflection process,
            # except with specific INCOSE-focused prompts
            reflection = self.reflection_chain.invoke(
                dict(
                    prompt=prompt_value.to_string(),
                    completion=completion,
                )
            )
            if re.search(r"\bLGTM\b", reflection) is not None:
                return self.parser.parse(completion)
            if not retry_number:
                log.debug(f"Completion:\n{completion}")
            completion = self.revision_chain.invoke(
                dict(
                    prompt=prompt_value.to_string(),
                    completion=completion,
                    reflection=reflection,
                )
            )

        return self.parser.parse(completion)


class HallucinationRefiner(ReflectionRefiner):
    def __init__(self, **kwargs):
        super().__init__(
            prompt_template_name="refinement/hallucination",
            **kwargs,
        )



import re
import subprocess  # nosec
from pathlib import Path
from tempfile import NamedTemporaryFile

from langchain_core.messages import BaseMessage

from janus.parsers.code_parser import CodeParser
from janus.parsers.parser import JanusParserException
from janus.utils.logger import create_logger

log = create_logger(__name__)


class UMLSyntaxParser(CodeParser):
    def _check_plantuml(self, text: str) -> None:
        original_text = text
        # Leading newlines can break the parser, remove them
        text = text.replace("\\n", "\n").strip()

        # Write the text to a temporary file (automatically deleted)
        file = NamedTemporaryFile()
        fname = file.name
        with open(fname, "w") as fin:
            fin.write(text)

        try:
            plantuml_path = Path.home().expanduser() / ".janus/lib/plantuml.jar"
            # NOTE: running subprocess with shell=False, added nosec to
            # label that we know risk exists
            res = subprocess.run(
                ["java", "-jar", plantuml_path, fname],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )  # nosec
            stdout = res.stdout.decode("utf-8")
            stderr = res.stderr.decode("utf-8")
        except FileNotFoundError:
            err_txt = (
                "Plant UML executable not found. Either choose a different parser"
                " or install with `bash scripts/install_plantuml.sh`. Java and"
                " graphviz are dependencies for the tool, they must also be installed."
            )
            log.error(err_txt)
            raise Exception(err_txt)

        # Check for bad outputs, raise JanusParserExceptions if so
        if "Error" in stderr or "Error" in stdout:
            err_txt = "Recieved UML parsing error(s)."

            line_nos = self._get_error_lines(stderr) + self._get_error_lines(stdout)
            lines = text.split("\n")
            for i in line_nos:
                i0 = max(0, i - 3)
                i1 = min(len(lines) - 1, i + 2)
                err_lines = [
                    f"> {lines[j]}" if j == i - 1 else f"  {lines[j]}"
                    for j in range(i0, i1)
                ]
                if i0:
                    err_lines.insert(0, "  ...")
                if i1 < (len(lines) - 1):
                    err_lines.append("  ...")

                err_txt += f"\nError located at line {i} must be fixed:\n"
                err_txt += "\n".join(err_lines)
            log.warning(err_txt)
            raise JanusParserException(original_text, err_txt)

        if "Warning" in stdout or "Warning" in stderr:
            err_txt = "Recieved UML parsing warning (often due to missing PLANTUML)."
            if stderr:
                err_txt += f"\nSTDERR:\n```\n{stderr.strip()}\n```\n"
            if stdout:
                err_txt += f"\nSTDOUT:\n```\n{stdout.strip()}\n```\n"

            log.warning(err_txt)
            raise JanusParserException(original_text, err_txt)

    def _get_error_lines(self, s: str) -> list[int]:
        return [int(x.group(1)) for x in re.finditer(r"Error line (\d+) in file:", s)]

    def _get_warns(self, s: str) -> list[str]:
        return [x.group() for x in re.finditer(r"Warning: (.*)\n", s)]

    def parse(self, text: str | BaseMessage) -> str:
        text = super().parse(text)
        self._check_plantuml(text)
        return text



from pathlib import Path
from typing import Optional

import click
import typer
from typing_extensions import Annotated

from janus.cli.constants import REFINERS
from janus.language.naive.registry import CUSTOM_SPLITTERS
from janus.utils.enums import LANGUAGES
from janus.utils.logger import create_logger

log = create_logger(__name__)


def translate(
    input_dir: Annotated[
        Path,
        typer.Option(
            "--input",
            "-i",
            help="The directory containing the source code to be translated. "
            "The files should all be in one flat directory.",
        ),
    ],
    source_lang: Annotated[
        str,
        typer.Option(
            "--source-language",
            "-s",
            help="The language of the source code.",
            click_type=click.Choice(sorted(LANGUAGES)),
        ),
    ],
    output_dir: Annotated[
        Path,
        typer.Option(
            "--output", "-o", help="The directory to store the translated code in."
        ),
    ],
    target_lang: Annotated[
        str,
        typer.Option(
            "--target-language",
            "-t",
            help="The desired output language to translate the source code to. The "
            "format can follow a 'language-version' syntax.  Use 'text' to get plaintext"
            "results as returned by the LLM. Examples: `python-3.10`, `mumps`, `java-10`,"
            "text.",
        ),
    ],
    llm_name: Annotated[
        str,
        typer.Option(
            "--llm",
            "-L",
            help="The custom name of the model set with 'janus llm add'.",
        ),
    ],
    failure_dir: Annotated[
        Optional[Path],
        typer.Option(
            "--failure-directory",
            "-f",
            help="The directory to store failure files during translation",
        ),
    ] = None,
    max_prompts: Annotated[
        int,
        typer.Option(
            "--max-prompts",
            "-m",
            help="The maximum number of times to prompt a model on one functional block "
            "before exiting the application. This is to prevent wasting too much money.",
        ),
    ] = 10,
    overwrite: Annotated[
        bool,
        typer.Option(
            "--overwrite/--preserve",
            help="Whether to overwrite existing files in the output directory",
        ),
    ] = False,
    skip_context: Annotated[
        bool,
        typer.Option(
            "--skip-context",
            help="Prompts will include any context information associated with source"
            " code blocks, unless this option is specified",
        ),
    ] = False,
    temp: Annotated[
        float,
        typer.Option("--temperature", "-T", help="Sampling temperature.", min=0, max=2),
    ] = 0.7,
    prompt_template: Annotated[
        str,
        typer.Option(
            "--prompt-template",
            "-p",
            help="Name of the Janus prompt template directory or "
            "path to a directory containing those template files.",
        ),
    ] = "simple",
    collection: Annotated[
        str,
        typer.Option(
            "--collection",
            "-c",
            help="If set, will put the translated result into a Chroma DB "
            "collection with the name provided.",
        ),
    ] = None,
    splitter_type: Annotated[
        str,
        typer.Option(
            "-S",
            "--splitter",
            help="Name of custom splitter to use",
            click_type=click.Choice(list(CUSTOM_SPLITTERS.keys())),
        ),
    ] = "file",
    refiner_types: Annotated[
        list[str],
        typer.Option(
            "-r",
            "--refiner",
            help="List of refiner types to use. Add -r for each refiner to use in\
                refinement chain",
            click_type=click.Choice(list(REFINERS.keys())),
        ),
    ] = ["JanusRefiner"],
    retriever_type: Annotated[
        str,
        typer.Option(
            "-R",
            "--retriever",
            help="Name of custom retriever to use",
            click_type=click.Choice(["active_usings", "language_docs"]),
        ),
    ] = None,
    max_tokens: Annotated[
        int,
        typer.Option(
            "--max-tokens",
            "-M",
            help="The maximum number of tokens the model will take in. "
            "If unspecificed, model's default max will be used.",
        ),
    ] = None,
):
    from janus.cli.constants import db_loc, get_collections_config
    from janus.converter.translate import Translator

    refiner_types = [REFINERS[r] for r in refiner_types]
    try:
        target_language, target_version = target_lang.split("-")
    except ValueError:
        target_language = target_lang
        target_version = None
    # make sure not overwriting input
    if source_lang.lower() == target_language.lower() and input_dir == output_dir:
        log.error("Output files would overwrite input! Aborting...")
        raise ValueError

    model_arguments = dict(temperature=temp)
    collections_config = get_collections_config()
    translator = Translator(
        model=llm_name,
        model_arguments=model_arguments,
        source_language=source_lang,
        target_language=target_language,
        target_version=target_version,
        max_prompts=max_prompts,
        max_tokens=max_tokens,
        prompt_template=prompt_template,
        db_path=db_loc,
        db_config=collections_config,
        splitter_type=splitter_type,
        refiner_types=refiner_types,
        retriever_type=retriever_type,
    )
    translator.translate(input_dir, output_dir, failure_dir, overwrite, collection)



from janus.llm.models_info import JanusModel
from janus.parsers.parser import JanusParser
from janus.refiners.refiner import ReflectionRefiner


class ALCFixUMLVariablesRefiner(ReflectionRefiner):
    def __init__(
        self,
        llm: JanusModel,
        parser: JanusParser,
        max_retries: int,
    ):
        super().__init__(
            llm=llm,
            parser=parser,
            max_retries=max_retries,
            prompt_template_name="refinement/uml/alc_fix_variables",
        )


class FixUMLConnectionsRefiner(ReflectionRefiner):
    def __init__(
        self,
        llm: JanusModel,
        parser: JanusParser,
        max_retries: int,
    ):
        super().__init__(
            llm=llm,
            parser=parser,
            max_retries=max_retries,
            prompt_template_name="refinement/uml/fix_connections",
        )



import click
import typer
from typing_extensions import Annotated

from janus.embedding.embedding_models_info import EmbeddingModelType

embedding = typer.Typer(
    help="Embedding model commands",
    add_completion=False,
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
)


@embedding.command("add", help="Add an embedding model config to janus")
def embedding_add(
    model_name: Annotated[
        str, typer.Argument(help="The user's custom name for the model")
    ],
    model_type: Annotated[
        str,
        typer.Option(
            "--type",
            "-t",
            help="The type of the model",
            click_type=click.Choice(list(val.value for val in EmbeddingModelType)),
        ),
    ] = "OpenAI",
):
    import json
    from pathlib import Path

    from pydantic import AnyHttpUrl

    from janus.embedding.embedding_models_info import (
        EMBEDDING_COST_PER_MODEL,
        EMBEDDING_MODEL_CONFIG_DIR,
        EMBEDDING_TOKEN_LIMITS,
    )

    if not EMBEDDING_MODEL_CONFIG_DIR.exists():
        EMBEDDING_MODEL_CONFIG_DIR.mkdir(parents=True)
    model_cfg = EMBEDDING_MODEL_CONFIG_DIR / f"{model_name}.json"
    if model_type in EmbeddingModelType.HuggingFaceInferenceAPI.values:
        hf = typer.style("HuggingFaceInferenceAPI", fg="yellow")
        url = typer.prompt(f"Enter the {hf} model's URL", type=str, value_proc=AnyHttpUrl)
        api_model_name = typer.prompt("Enter the model's name", type=str, default="")
        api_key = typer.prompt("Enter the API key", type=str, default="")
        max_tokens = typer.prompt(
            "Enter the model's maximum tokens", default=8191, type=int
        )
        in_cost = typer.prompt("Enter the cost per input token", default=0, type=float)
        out_cost = typer.prompt("Enter the cost per output token", default=0, type=float)
        params = dict(
            model_name=api_model_name,
            api_key=api_key,
        )
        cfg = {
            "model_type": model_type,
            "model_identifier": str(url),
            "model_args": params,
            "token_limit": max_tokens,
            "model_cost": {"input": in_cost, "output": out_cost},
        }
    elif model_type in EmbeddingModelType.HuggingFaceLocal.values:
        hf = typer.style("HuggingFace", fg="yellow")
        model_id = typer.prompt(
            f"Enter the {hf} model ID",
            default="sentence-transformers/all-MiniLM-L6-v2",
            type=str,
        )
        cache_folder = str(
            Path(
                typer.prompt(
                    "Enter the model's cache folder",
                    default=EMBEDDING_MODEL_CONFIG_DIR / "cache",
                    type=str,
                )
            )
        )
        max_tokens = typer.prompt(
            "Enter the model's maximum tokens", default=8191, type=int
        )
        params = dict(
            cache_folder=str(cache_folder),
        )
        cfg = {
            "model_type": model_type,
            "model_identifier": model_id,
            "model_args": params,
            "token_limit": max_tokens,
            "model_cost": {"input": 0, "output": 0},
        }
    elif model_type in EmbeddingModelType.OpenAI.values:
        available_models = list(EMBEDDING_COST_PER_MODEL.keys())

        open_ai = typer.style("OpenAI", fg="green")
        prompt = f"Enter the {open_ai} model name"

        model_name = typer.prompt(
            prompt,
            default="text-embedding-3-small",
            type=click.types.Choice(available_models),
            show_choices=False,
        )
        params = dict(
            model=model_name,
        )
        max_tokens = EMBEDDING_TOKEN_LIMITS[model_name]
        model_cost = EMBEDDING_COST_PER_MODEL[model_name]
        cfg = {
            "model_type": model_type,
            "model_identifier": model_name,
            "model_args": params,
            "token_limit": max_tokens,
            "model_cost": model_cost,
        }
    else:
        raise ValueError(f"Unknown model type {model_type}")
    with open(model_cfg, "w") as f:
        json.dump(cfg, f, indent=2)
    print(f"Model config written to {model_cfg}")



from typing import Any

from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompt_values import PromptValue
from langchain_core.runnables import RunnableSerializable

from janus.llm.models_info import MODEL_PROMPT_ENGINES, JanusModel
from janus.parsers.parser import JanusParser
from janus.refiners.refiner import JanusRefiner


class FormatRefiner(JanusRefiner):
    format_chain: RunnableSerializable

    def __init__(
        self,
        llm: JanusModel,
        parser: JanusParser,
        max_retries: int,
        prompt_template_name: str,
    ):
        format_prompt = MODEL_PROMPT_ENGINES[llm.short_model_id](
            source_language="text",
            prompt_template=prompt_template_name,
        ).prompt
        format_chain = format_prompt | llm | StrOutputParser()
        super().__init__(
            format_chain=format_chain, parser=parser, max_retries=max_retries
        )

    def parse_completion(
        self, completion: str, prompt_value: PromptValue, **kwargs
    ) -> Any:
        completion = self.format_chain.invoke(
            dict(completion=completion, prompt=prompt_value.to_string())
        )
        return self.parser.parse(completion)


class CodeFormatRefiner(FormatRefiner):
    def __init__(self, llm: JanusModel, parser: JanusParser, max_retries: int):
        super().__init__(llm, parser, max_retries, "refinement/format/code_format")


class RequirementsFormatRefiner(FormatRefiner):
    def __init__(self, llm: JanusModel, parser: JanusParser, max_retries: int):
        super().__init__(
            llm, parser, max_retries, "refinement/format/requirements_format"
        )



import json
from pathlib import Path

import janus.refiners.format
import janus.refiners.refiner
import janus.refiners.uml

homedir = Path.home().expanduser()

janus_dir = homedir / ".janus"
if not janus_dir.exists():
    janus_dir.mkdir(parents=True)

db_file = janus_dir / ".db"
if not db_file.exists():
    with open(db_file, "w") as f:
        f.write(str(janus_dir / "chroma.db"))

with open(db_file, "r") as f:
    db_loc = f.read()

collections_config_file = Path(db_loc) / "collections.json"


def get_subclasses(cls):
    return set(cls.__subclasses__()).union(
        set(s for c in cls.__subclasses__() for s in get_subclasses(c))
    )


REFINER_TYPES = get_subclasses(janus.refiners.refiner.JanusRefiner).union(
    {janus.refiners.refiner.JanusRefiner}
)
REFINERS = {r.__name__: r for r in REFINER_TYPES}


def get_collections_config():
    if collections_config_file.exists():
        with open(collections_config_file, "r") as f:
            config = json.load(f)
    else:
        config = {}
    return config



import json
import re
from copy import deepcopy

from langchain_core.runnables import Runnable, RunnableLambda, RunnableParallel

from janus.converter.converter import Converter
from janus.language.block import TranslatedCodeBlock
from janus.language.combine import JsonCombiner
from janus.parsers.eval_parsers.incose_parser import IncoseParser
from janus.parsers.eval_parsers.inline_comment_parser import InlineCommentParser
from janus.parsers.parser import JanusParserException
from janus.utils.logger import create_logger

log = create_logger(__name__)


class Evaluator(Converter):
    """Evaluator

    A class that performs an LLM self evaluation"
    "on an input target, with an associated prompt.

    Current valid evaluation types:
    ['incose', 'comments']

    """

    def __init__(self, **kwargs) -> None:
        """Initialize the Evaluator class

        Arguments:
            model: The LLM to use for translation. If an OpenAI model, the
                `OPENAI_API_KEY` environment variable must be set.
            model_arguments: Additional arguments to pass to the LLM constructor.
            max_prompts: The maximum number of prompts to try before giving up.
        """
        super().__init__(**kwargs)
        self._combiner = JsonCombiner()
        self._load_parameters()


class RequirementEvaluator(Evaluator):
    """INCOSE Requirement Evaluator

    A class that performs an LLM self evaluation on an input target,
    with an associated prompt.

    The evaluation prompts are for Incose Evaluations

    """

    def __init__(self, eval_items_per_request: int | None = None, **kwargs) -> None:
        """Initialize the Evaluator class

        Arguments:
            model: The LLM to use for translation. If an OpenAI model, the
                `OPENAI_API_KEY` environment variable must be set.
            model_arguments: Additional arguments to pass to the LLM constructor.
            max_prompts: The maximum number of prompts to try before giving up.
        """
        super().__init__(**kwargs)
        self.eval_items_per_request = eval_items_per_request
        self._parser = IncoseParser()
        self.set_prompt("eval_prompts/incose")

    def _input_runnable(self) -> Runnable:
        def _get_code(json_text: str) -> str:
            return json.loads(json_text)["code"]

        def _get_reqs(json_text: str) -> str:
            return json.dumps(json.loads(json_text)["requirements"])

        return RunnableLambda(self._parser.parse_input) | RunnableParallel(
            SOURCE_CODE=_get_code,
            REQUIREMENTS=_get_reqs,
            context=self._retriever,
        )

    def _add_translation(self, block: TranslatedCodeBlock):
        if block.translated:
            return

        if block.original.text is None:
            block.translated = True
            return

        if self.eval_items_per_request is None:
            return super()._add_translation(block)

        input_obj = json.loads(block.original.text)
        requirements = input_obj.get("requirements", [])

        if not requirements:
            log.debug(f"[{block.name}] Skipping empty block")
            block.translated = True
            block.text = None
            block.complete = True
            return

        # For some reason requirements objects are in nested lists?
        while isinstance(requirements[0], list):
            requirements = [r for lst in requirements for r in lst]

        if len(requirements) <= self.eval_items_per_request:
            input_obj["requirements"] = requirements
            block.original.text = json.dumps(input_obj)
            return super()._add_translation(block)

        block.processing_time = 0
        block.cost = 0
        obj = {}
        for i in range(0, len(requirements), self.eval_items_per_request):
            # Build a new TranslatedBlock using the new working text
            working_requirements = requirements[i : i + self.eval_items_per_request]
            working_copy = deepcopy(block.original)
            working_obj = json.loads(working_copy.text)  # type: ignore
            working_obj["requirements"] = working_requirements
            working_copy.text = json.dumps(working_obj)
            working_block = TranslatedCodeBlock(working_copy, self._target_language)

            # Run the LLM on the working text
            try:
                super()._add_translation(working_block)
            except JanusParserException as e:
                block.text += "\n==============\n" + working_block.text
                block.tokens = self._llm.get_num_tokens(block.text)
                raise e
            finally:
                # Update metadata to include for all runs
                block.num_requests += working_block.num_requests
                block.cost += working_block.cost
                block.processing_time += working_block.processing_time
                block.request_input_tokens += working_block.request_input_tokens
                block.request_output_tokens += working_block.request_output_tokens

            # Update the output text to merge this section's output in
            obj.update(json.loads(working_block.text))
            # intermediate result of block,
            # will be overwritten if file completes successfully
            block.text = json.dumps(obj)

        block.text = json.dumps(obj)
        block.tokens = self._llm.get_num_tokens(block.text)
        block.translated = True

        log.debug(
            f"[{block.name}] Output code:\n{json.dumps(json.loads(block.text), indent=2)}"
        )


class InlineCommentEvaluator(Evaluator):
    """Inline Comment Evaluator

    A class that performs an LLM self evaluation on inline comments,
    with an associated prompt.
    """

    def __init__(self, eval_items_per_request: int | None = None, **kwargs) -> None:
        """Initialize the Evaluator class

        Arguments:
            model: The LLM to use for translation. If an OpenAI model, the
                `OPENAI_API_KEY` environment variable must be set.
            model_arguments: Additional arguments to pass to the LLM constructor.
            max_prompts: The maximum number of prompts to try before giving up.
        """
        super().__init__(**kwargs)
        self._combiner = JsonCombiner()
        self._load_parameters()
        self._parser = InlineCommentParser()
        self.set_prompt("eval_prompts/inline_comments")
        self.eval_items_per_request = eval_items_per_request

    def _add_translation(self, block: TranslatedCodeBlock):
        if block.translated:
            return

        if block.original.text is None:
            block.translated = True
            return

        if self.eval_items_per_request is None:
            return super()._add_translation(block)

        comment_pattern = r"<(?:INLINE|BLOCK)_COMMENT \w{8}>.*$"
        comments = list(
            re.finditer(comment_pattern, block.original.text, flags=re.MULTILINE)
        )

        if not comments:
            log.info(f"[{block.name}] Skipping commentless block")
            block.translated = True
            block.text = None
            block.complete = True
            return

        if len(comments) <= self.eval_items_per_request:
            return super()._add_translation(block)

        comment_group_indices = list(range(0, len(comments), self.eval_items_per_request))
        log.debug(
            f"[{block.name}] Block contains more than {self.eval_items_per_request}"
            f" comments, splitting {len(comments)} comments into"
            f" {len(comment_group_indices)} groups"
        )

        block.processing_time = 0
        block.cost = 0
        block.retries = 0
        obj = {}
        for i in range(0, len(comments), self.eval_items_per_request):
            # Split the text into the section containing comments of interest,
            #  all the text prior to those comments, and all the text after them
            working_comments = comments[i : i + self.eval_items_per_request]
            start_idx = working_comments[0].start()
            end_idx = working_comments[-1].end()
            prefix = block.original.text[:start_idx]
            keeper = block.original.text[start_idx:end_idx]
            suffix = block.original.text[end_idx:]

            # Strip all comment placeholders outside of the section of interest
            prefix = re.sub(comment_pattern, "", prefix, flags=re.MULTILINE)
            suffix = re.sub(comment_pattern, "", suffix, flags=re.MULTILINE)

            # Build a new TranslatedBlock using the new working text
            working_copy = deepcopy(block.original)
            working_copy.text = prefix + keeper + suffix
            working_block = TranslatedCodeBlock(working_copy, self._target_language)

            # Run the LLM on the working text
            super()._add_translation(working_block)

            # Update metadata to include for all runs
            block.retries += working_block.retries
            block.cost += working_block.cost
            block.processing_time += working_block.processing_time

            # Update the output text to merge this section's output in
            obj.update(json.loads(working_block.text))

        block.text = json.dumps(obj)
        block.tokens = self._llm.get_num_tokens(block.text)
        block.translated = True

        log.debug(
            f"[{block.name}] Output code:\n{json.dumps(json.loads(block.text), indent=2)}"
        )



from janus.language.alc.alc import AlcListingSplitter, AlcSplitter
from janus.language.mumps.mumps import MumpsSplitter
from janus.language.naive.registry import register_splitter
from janus.language.splitter import Splitter
from janus.language.treesitter import TreeSitterSplitter
from janus.utils.enums import LANGUAGES
from janus.utils.logger import create_logger

log = create_logger(__name__)


@register_splitter("ast-flex")
def get_flexible_ast(language: str, **kwargs) -> Splitter:
    """Get a flexible AST splitter for the given language.

    Arguments:
        language: The language to get the splitter for.

    Returns:
        A flexible AST splitter for the given language.
    """
    kwargs.update(protected_node_types=())
    if language == "ibmhlasm":
        return AlcSplitter(**kwargs)
    elif language == "mumps":
        return MumpsSplitter(**kwargs)
    else:
        return TreeSitterSplitter(language=language, **kwargs)


@register_splitter("ast-strict")
def get_strict_ast(language: str, prune_unprotected=True, **kwargs) -> Splitter:
    """Get a strict AST splitter for the given language.

    The strict splitter will only return nodes that are of a functional type.

    Arguments:
        language: The language to get the splitter for.

    Returns:
        A strict AST splitter for the given language.
    """
    kwargs.update(
        protected_node_types=LANGUAGES[language]["functional_node_types"],
        prune_unprotected=prune_unprotected,
    )
    if language == "ibmhlasm":
        return AlcSplitter(**kwargs)
    elif language == "mumps":
        return MumpsSplitter(**kwargs)
    else:
        return TreeSitterSplitter(language=language, **kwargs)


@register_splitter("ast-strict-listing")
def get_strict_listing_ast(language: str, **kwargs) -> Splitter:
    """Get a strict AST splitter for the given language. This splitter is intended for
    use with IBM HLASM.

    The strict splitter will only return nodes that are of a functional type.

    Arguments:
        language: The language to get the splitter for.

    Returns:
        A strict AST splitter for the given language.
    """
    kwargs.update(
        protected_node_types=LANGUAGES[language]["functional_node_types"],
        prune_unprotected=True,
    )
    if language == "ibmhlasm":
        return AlcListingSplitter(**kwargs)
    else:
        log.warning("Listing splitter is only intended for use with IBMHLASM!")
        return TreeSitterSplitter(language=language, **kwargs)


@register_splitter("ast-flex-listing")
def get_flexible_listing_ast(language: str, **kwargs) -> Splitter:
    """Get a flexible AST splitter for the given language. This splitter is intended for
    use with IBM HLASM.

    Arguments:
        language: The language to get the splitter for.

    Returns:
        A flexible AST splitter for the given language.
    """
    if language == "ibmhlasm":
        return AlcListingSplitter(**kwargs)
    else:
        log.warning("Listing splitter is only intended for use with IBMHLASM!")
        return TreeSitterSplitter(language=language, **kwargs)



from pathlib import Path
from typing import Optional

import typer
from typing_extensions import Annotated

from janus.cli.constants import db_loc, janus_dir

db = typer.Typer(
    help="Database commands",
    add_completion=False,
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
)


@db.command("init", help="Connect to or create a database.")
def db_init(
    path: Annotated[
        str, typer.Option("--path", "-p", help="The path to the database file.")
    ] = str(janus_dir / "chroma.db"),
    url: Annotated[
        str,
        typer.Option(
            "--url",
            "-u",
            help="The URL of the database if the database is running externally.",
        ),
    ] = "",
) -> None:
    import os

    from rich import print

    from janus.cli.constants import db_file
    from janus.embedding.database import ChromaEmbeddingDatabase

    global db_loc
    if url != "":
        print(f"Pointing to Chroma DB at {url}")
        with open(db_file, "w") as f:
            f.write(url)
        db_loc = url
    else:
        path = os.path.abspath(path)
        print(f"Setting up Chroma DB at {path}")
        with open(db_file, "w") as f:
            f.write(path)
        db_loc = path
    global embedding_db
    embedding_db = ChromaEmbeddingDatabase(db_loc)


@db.command("status", help="Print current database location.")
def db_status():
    from rich import print

    print(f"Chroma DB currently pointing to {db_loc}")


@db.command(
    "ls",
    help="List the current database's collections. Or supply a collection name to list "
    "information about its contents.",
)
def db_ls(
    collection_name: Annotated[
        Optional[str], typer.Argument(help="The name of the collection.")
    ] = None,
    peek: Annotated[
        Optional[int],
        typer.Option("--peek", "-p", help="Peek at N entries for a specific collection."),
    ] = None,
) -> None:
    """List the current database's collections"""
    from rich import print

    from janus.embedding.database import ChromaEmbeddingDatabase

    if peek is not None and collection_name is None:
        print(
            "\n[bold red]Cannot peek at all collections. Please specify a "
            "collection by name.[/bold red]"
        )
        return
    db = ChromaEmbeddingDatabase(db_loc)
    from janus.embedding.collections import Collections

    collections = Collections(db)
    collection_list = collections.get(collection_name)
    for collection in collection_list:
        print(
            f"\n[bold underline]Collection[/bold underline]: "
            f"[bold salmon1]{collection.name}[/bold salmon1]"
        )
        print(f"  ID: {collection.id}")
        print(f"  Metadata: {collection.metadata}")
        print(f"  Tenant: [green]{collection.tenant}[/green]")
        print(f"  Database: [green]{collection.database}[/green]")
        print(f"  Length: {collection.count()}")
        if peek:
            entry = collection.peek(peek)
            entry["embeddings"] = entry["embeddings"][0][:2] + ["..."]
            if peek == 1:
                print("  [bold]Peeking at first entry[/bold]:")
            else:
                print(f"  [bold]Peeking at first {peek} entries[/bold]:")
            print(entry)
        print()


@db.command("add", help="Add a collection to the current database.")
def db_add(
    collection_name: Annotated[str, typer.Argument(help="The name of the collection.")],
    model_name: Annotated[str, typer.Argument(help="The name of the embedding model.")],
    input_dir: Annotated[
        str,
        typer.Option(
            "--input",
            "-i",
            help="The directory containing the source code to be added.",
        ),
    ] = "./",
    input_lang: Annotated[
        str, typer.Option("--language", "-l", help="The language of the source code.")
    ] = "python",
    max_tokens: Annotated[
        int,
        typer.Option(
            "--max-tokens",
            "-m",
            help="The maximum number of tokens for each chunk of input source code.",
        ),
    ] = 4096,
) -> None:
    """Add a collection to the database

    Arguments:
        collection_name: The name of the collection to add
        model_name: The name of the embedding model to use
        input_dir: The directory containing the source code to be added
        input_lang: The language of the source code
        max_tokens: The maximum number of tokens for each chunk of input source code
    """
    # TODO: import factory
    import json
    from pathlib import Path

    from rich.console import Console

    from janus.cli.constants import collections_config_file, get_collections_config
    from janus.embedding.vectorize import ChromaDBVectorizer
    from janus.language.binary import BinarySplitter
    from janus.language.mumps import MumpsSplitter
    from janus.language.naive.registry import CUSTOM_SPLITTERS
    from janus.language.treesitter import TreeSitterSplitter
    from janus.utils.enums import LANGUAGES

    console = Console()

    added_to = _check_collection(collection_name, input_dir)
    collections_config = get_collections_config()

    with console.status(
        f"Adding collection: [bold salmon]{collection_name}[/bold salmon]",
        spinner="arrow3",
    ):
        vectorizer_factory = ChromaDBVectorizer()
        vectorizer = vectorizer_factory.create_vectorizer(
            path=db_loc, config=collections_config
        )
        vectorizer.get_or_create_collection(collection_name, model_name=model_name)
        input_dir = Path(input_dir)
        suffixes = [f".{ext}" for ext in LANGUAGES[input_lang]["suffixes"]]
        input_paths = [file for ext in suffixes for file in input_dir.rglob(f"**/*{ext}")]

        if input_lang in CUSTOM_SPLITTERS:
            if input_lang == "mumps":
                splitter = MumpsSplitter(
                    max_tokens=max_tokens,
                )
            elif input_lang == "binary":
                splitter = BinarySplitter(
                    max_tokens=max_tokens,
                )
        else:
            splitter = TreeSitterSplitter(
                language=input_lang,
                max_tokens=max_tokens,
            )
        for input_path in input_paths:
            input_block = splitter.split(input_path)
            vectorizer.add_nodes_recursively(
                input_block,
                collection_name,
                input_path.name,
            )
    total_files = len(
        [path for path in Path.glob(input_dir, "**/*") if not path.is_dir()]
    )
    if added_to:
        print(
            f"\nAdded to [bold salmon1]{collection_name}[/bold salmon1]:\n"
            f"  Embedding Model: [green]{model_name}[/green]\n"
            f"  Input Directory: {input_dir.absolute()}\n"
            f"  {input_lang} [green]{suffixes}[/green] Files: "
            f"{len(input_paths)}\n"
            "  Other Files (skipped): "
            f"{total_files - len(input_paths)}\n"
        )
    else:
        print(
            f"\nCreated [bold salmon1]{collection_name}[/bold salmon1]:\n"
            f"  Embedding Model: '{model_name}'\n"
            f"  Input Directory: {input_dir.absolute()}\n"
            f"  {input_lang} [green]{suffixes}[/green] Files: "
            f"{len(input_paths)}\n"
            "  Other Files (skipped): "
            f"{total_files - len(input_paths)}\n"
        )
    with open(collections_config_file, "w") as f:
        json.dump(vectorizer.config, f, indent=2)


@db.command(
    "rm",
    help="Remove a collection from the database.",
)
def db_rm(
    collection_name: Annotated[str, typer.Argument(help="The name of the collection.")],
    confirm: Annotated[
        bool,
        typer.Option(
            "--yes",
            "-y",
            help="Confirm the removal of the collection.",
        ),
    ],
) -> None:
    """Remove a collection from the database

    Arguments:
        collection_name: The name of the collection to remove
    """
    from rich.prompt import Confirm

    from janus.embedding.collections import Collections
    from janus.embedding.database import ChromaEmbeddingDatabase

    if not confirm:
        delete = Confirm.ask(
            f"\nAre you sure you want to [bold red]remove[/bold red] "
            f"[bold salmon1]{collection_name}[/bold salmon1]?",
        )
    else:
        delete = True
    if not delete:
        raise typer.Abort()
    db = ChromaEmbeddingDatabase(db_loc)
    collections = Collections(db)
    collections.delete(collection_name)
    print(
        f"[bold red]Removed[/bold red] collection "
        f"[bold salmon1]{collection_name}[/bold salmon1]"
    )


def _check_collection(collection_name: str, input_dir: str | Path) -> bool:
    from chromadb.errors import InvalidCollectionException

    from janus.embedding.collections import Collections
    from janus.embedding.database import ChromaEmbeddingDatabase

    db = ChromaEmbeddingDatabase(db_loc)
    collections = Collections(db)
    added_to = False
    try:
        collections.get(collection_name)
        # confirm_add = Confirm.ask(
        #     f"\nCollection [bold salmon1]{collection_name}[/bold salmon1] exists. Are "
        #     "you sure you want to update it with the contents of"
        #     f"[bold green]{input_dir}[/bold green]?"
        # )
        added_to = True
        # if not confirm_add:
        #     raise typer.Abort()
    except InvalidCollectionException:
        pass
    return added_to



from pathlib import Path

from janus.converter.converter import Converter
from janus.language.block import TranslatedCodeBlock
from janus.parsers.partition_parser import PartitionParser
from janus.utils.logger import create_logger

log = create_logger(__name__)


class Partitioner(Converter):
    def __init__(self, partition_token_limit: int, **kwargs):
        super().__init__(**kwargs)
        self.set_prompt("partition")
        self._load_model()
        self._parser = PartitionParser(
            token_limit=partition_token_limit,
            model=self._llm,
        )
        self._target_language = self._source_language
        self._target_suffix = self._source_suffix
        self._load_parameters()

    def _save_to_file(self, block: TranslatedCodeBlock, out_path: Path) -> None:
        output_str = self._parser.parse_combined_output(block.complete_text)
        out_path.parent.mkdir(parents=True, exist_ok=True)
        out_path.write_text(output_str, encoding="utf-8")



import json
import random
import uuid

from langchain.output_parsers import PydanticOutputParser
from langchain_core.exceptions import OutputParserException
from langchain_core.language_models import BaseLanguageModel
from langchain_core.messages import BaseMessage
from langchain_core.pydantic_v1 import BaseModel, Field

from janus.language.block import CodeBlock
from janus.parsers.parser import JanusParser, JanusParserException
from janus.utils.logger import create_logger

log = create_logger(__name__)
RNG = random.Random()


class PartitionObject(BaseModel):
    reasoning: str = Field(
        description="An explanation for why the code should be split at this point"
    )
    location: str = Field(
        description="The 8-character line label which should start a new chunk"
    )


class PartitionList(BaseModel):
    __root__: list[PartitionObject] = Field(
        description=(
            "A list of appropriate split points, each with a `reasoning` field "
            "that explains a justification for splitting the code at that point, "
            "and a `location` field which is simply the 8-character line ID. "
            "The `reasoning` field should always be included first."
        )
    )


# The following IDs appear in the prompt example. If the LLM produces them,
#  they should be ignored
EXAMPLE_IDS = {
    "0d2f4f8d",
    "def2a953",
    "75315253",
    "e7f928da",
    "1781b2a9",
    "2fe21e27",
    "9aef6179",
    "6061bd82",
    "22bd0c30",
    "5d85e19e",
    "06027969",
    "91b722fb",
    "4b3f79be",
    "k57w964a",
    "51638s96",
    "065o6q32",
    "j5q6p852",
}


class PartitionParser(JanusParser, PydanticOutputParser):
    token_limit: int
    model: BaseLanguageModel
    lines: list[str] = []
    line_id_to_index: dict[str, int] = {}

    def __init__(self, token_limit: int, model: BaseLanguageModel):
        PydanticOutputParser.__init__(
            self,
            pydantic_object=PartitionList,
            model=model,
            token_limit=token_limit,
        )

    def parse_input(self, block: CodeBlock) -> str:
        code = str(block.text)
        RNG.seed(code)

        self.lines = code.split("\n")

        # Generate a unique ID for each line (ensure they are unique)
        line_ids = set()
        while len(line_ids) < len(self.lines):
            line_id = str(uuid.UUID(int=RNG.getrandbits(128), version=4))[:8]
            if line_id in EXAMPLE_IDS:
                continue
            line_ids.add(line_id)

        # Prepend each line with the corresponding ID, save the mapping
        self.line_id_to_index = {lid: i for i, lid in enumerate(line_ids)}
        processed = "\n".join(
            f"{line_id}\t{self.lines[i]}" for line_id, i in self.line_id_to_index.items()
        )
        return processed

    def parse(self, text: str | BaseMessage) -> str:
        if isinstance(text, BaseMessage):
            text = str(text.content)
        original_text = text

        # Strip everything outside the JSON object
        begin, end = text.find("["), text.rfind("]")
        text = text[begin : end + 1]

        try:
            out: PartitionList = super().parse(text)
        except (OutputParserException, json.JSONDecodeError):
            log.debug(f"Invalid JSON object. Output:\n{text}")
            raise

        # Get partition locations, discard reasoning
        partition_locations = {partition.location for partition in out.__root__}

        # Ignore IDs from the example input
        partition_locations.difference_update(EXAMPLE_IDS)

        # Locate any invalid line IDs, raise exception if any found
        invalid_splits = partition_locations.difference(self.line_id_to_index)
        if invalid_splits:
            err_msg = (
                f"{len(invalid_splits)} line ID(s) not found in input: "
                + ", ".join(invalid_splits)
            )
            log.warning(err_msg)
            raise JanusParserException(original_text, err_msg)

        # Map line IDs to indices (so they can be sorted and lines indexed)
        index_to_line_id = {0: "START", None: "END"}
        split_points = {0}
        for partition in partition_locations:
            index = self.line_id_to_index[partition]
            index_to_line_id[index] = partition
            split_points.add(index)

        # Get partition start/ends, chunks, chunk lengths
        split_points = sorted(split_points) + [None]
        partition_indices = list(zip(split_points, split_points[1:]))
        partition_points = [
            (index_to_line_id[i0], index_to_line_id[i1]) for i0, i1 in partition_indices
        ]
        chunks = ["\n".join(self.lines[i0:i1]) for i0, i1 in partition_indices]
        chunk_tokens = list(map(self.model.get_num_tokens, chunks))

        # Collect any chunks that exceed token limit
        oversized_indices: list[int] = [
            i for i, n in enumerate(chunk_tokens) if n > self.token_limit
        ]
        if oversized_indices:
            data = list(zip(partition_points, chunks, chunk_tokens))
            data = [data[i] for i in oversized_indices]

            problem_points = "\n".join(
                [
                    f"{i0} to {i1} ({t / self.token_limit:.1f}x maximum length)"
                    for (i0, i1), _, t in data
                ]
            )
            log.warning(f"Found {len(data)} oversized chunks:\n{problem_points}")
            log.debug(
                "Oversized chunks:\n"
                + "\n#############\n".join(chunk for _, chunk, _ in data)
            )
            raise JanusParserException(
                original_text,
                f"The following segments are too long and must be "
                f"further subdivided:\n{problem_points}",
            )

        return "\n<JANUS_PARTITION>\n".join(chunks)



from typing import List

from langchain_core.documents import Document
from langchain_core.output_parsers import StrOutputParser
from langchain_core.retrievers import BaseRetriever
from langchain_core.runnables import Runnable, RunnableConfig

from janus.language.block import CodeBlock
from janus.llm.models_info import MODEL_PROMPT_ENGINES, JanusModel
from janus.utils.logger import create_logger
from janus.utils.pdf_docs_reader import PDFDocsReader

log = create_logger(__name__)


class JanusRetriever(Runnable):
    def __init__(self) -> None:
        super().__init__()

    def invoke(
        self, input: CodeBlock, config: RunnableConfig | None = None, **kwargs
    ) -> dict:
        kwargs.update(context=self.get_context(input))
        return kwargs

    def get_context(self, code_block: CodeBlock) -> str:
        return ""


class ActiveUsingsRetriever(JanusRetriever):
    def get_context(self, code_block: CodeBlock) -> str:
        context = "\n".join(
            f"{context_tag}: {context}"
            for context_tag, context in code_block.context_tags.items()
        )
        return f"You may use the following additional context: {context}"


class TextSearchRetriever(JanusRetriever):
    retriever: BaseRetriever

    def __init__(self, retriever: BaseRetriever):
        super().__init__()
        self.retriever = retriever

    def get_context(self, code_block: CodeBlock) -> str:
        if code_block.text is None:
            return ""
        docs = self.retriever.invoke(code_block.text)
        context = "\n\n".join(doc.page_content for doc in docs)
        return f"You may use the following additional context: {context}"


class LanguageDocsRetriever(JanusRetriever):
    def __init__(
        self,
        llm: JanusModel,
        language_name: str,
        prompt_template_name: str = "retrieval/language_docs",
    ):
        super().__init__()
        self.llm: JanusModel = llm
        self.language: str = language_name

        self.PDF_reader = PDFDocsReader(
            language=self.language,
        )

        language_docs_prompt = MODEL_PROMPT_ENGINES[self.llm.short_model_id](
            source_language=self.language,
            prompt_template=prompt_template_name,
        ).prompt

        parser: StrOutputParser = StrOutputParser()
        self.chain = language_docs_prompt | self.llm | parser

    def get_context(self, code_block: CodeBlock) -> str:
        functionality_to_reference: str = self.chain.invoke(
            dict({"SOURCE_CODE": code_block.text, "SOURCE_LANGUAGE": self.language})
        )
        if functionality_to_reference == "NODOCS":
            log.debug("No Opcodes requested from language docs retriever.")
            return ""
        else:
            functionality_to_reference: List = functionality_to_reference.split(", ")
            log.debug(
                f"List of opcodes requested by language docs retriever"
                f"to search the {self.language} "
                f"docs for: {functionality_to_reference}"
            )

            docs: List[Document] = self.PDF_reader.search_language_reference(
                functionality_to_reference
            )
            context = "\n\n".join(doc.page_content for doc in docs)
            if context:
                return (
                    f"You may reference the following excerpts from the {self.language} "
                    f"language documentation: {context}"
                )
            else:
                return ""



import click
import typer
from typing_extensions import Annotated

from janus.llm.models_info import MODEL_TYPE_CONSTRUCTORS

llm = typer.Typer(
    help="LLM commands",
    add_completion=False,
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
)


@llm.command("add", help="Add a model config to janus")
def llm_add(
    model_name: Annotated[
        str, typer.Argument(help="The user's custom name of the model")
    ],
    model_type: Annotated[
        str,
        typer.Option(
            "--type",
            "-t",
            help="The type of the model",
            click_type=click.Choice(sorted(list(MODEL_TYPE_CONSTRUCTORS.keys()))),
        ),
    ] = "Azure",
):
    import json

    from janus.llm.models_info import (
        COST_PER_1K_TOKENS,
        MODEL_CONFIG_DIR,
        MODEL_ID_TO_LONG_ID,
        TOKEN_LIMITS,
        azure_models,
        bedrock_models,
        openai_models,
    )

    if not MODEL_CONFIG_DIR.exists():
        MODEL_CONFIG_DIR.mkdir(parents=True)
    model_cfg = MODEL_CONFIG_DIR / f"{model_name}.json"
    if model_type == "HuggingFace":
        url = typer.prompt("Enter the model's URL")
        max_tokens = typer.prompt(
            "Enter the model's maximum tokens", default=4096, type=int
        )
        in_cost = typer.prompt("Enter the cost per input token", default=0, type=float)
        out_cost = typer.prompt("Enter the cost per output token", default=0, type=float)
        params = dict(
            inference_server_url=url,
            max_new_tokens=max_tokens,
            top_k=10,
            top_p=0.95,
            typical_p=0.95,
            temperature=0.01,
            repetition_penalty=1.03,
            timeout=240,
        )
        cfg = {
            "model_type": model_type,
            "model_args": params,
            "token_limit": max_tokens,
            "model_cost": {"input": in_cost, "output": out_cost},
            "input_token_proportion": 0.4,
        }
    elif model_type == "HuggingFaceLocal":
        model_id = typer.prompt("Enter the model ID")
        task = typer.prompt("Enter the task")
        max_tokens = typer.prompt(
            "Enter the model's maximum tokens", default=4096, type=int
        )
        in_cost = 0
        out_cost = 0
        params = {"model_id": model_id, "task": task}
        cfg = {
            "model_type": model_type,
            "model_args": params,
            "token_limit": max_tokens,
            "model_cost": {"input": in_cost, "output": out_cost},
            "input_token_proportion": 0.4,
        }
    elif model_type == "OpenAI":
        print("DEPRECATED: Use 'Azure' instead. CTRL+C to exit.")
        model_id = typer.prompt(
            "Enter the model ID (list model IDs with `janus llm ls -a`)",
            default="gpt-4o",
            type=click.Choice(openai_models),
            show_choices=False,
        )
        params = dict(
            model_name=model_name,
            temperature=0.7,
            n=1,
        )
        max_tokens = TOKEN_LIMITS[model_name]
        model_cost = COST_PER_1K_TOKENS[model_name]
        cfg = {
            "model_type": model_type,
            "model_id": model_id,
            "model_args": params,
            "token_limit": max_tokens,
            "model_cost": model_cost,
            "input_token_proportion": 0.4,
        }
    elif model_type == "Azure":
        model_id = typer.prompt(
            "Enter the model ID (list model IDs with `janus llm ls -a`)",
            default="gpt-4o",
            type=click.Choice(azure_models),
            show_choices=False,
        )
        params = dict(
            # Azure uses the "azure_deployment" key for what we're calling "long_model_id"
            azure_deployment=MODEL_ID_TO_LONG_ID[model_id],
            temperature=0.7,
            n=1,
        )
        max_tokens = TOKEN_LIMITS[MODEL_ID_TO_LONG_ID[model_id]]
        model_cost = COST_PER_1K_TOKENS[MODEL_ID_TO_LONG_ID[model_id]]
        cfg = {
            "model_type": model_type,
            "model_id": model_id,
            "model_args": params,
            "token_limit": max_tokens,
            "model_cost": model_cost,
            "input_token_proportion": 0.4,
        }
    elif model_type == "BedrockChat" or model_type == "Bedrock":
        model_id = typer.prompt(
            "Enter the model ID (list model IDs with `janus llm ls -a`)",
            default="bedrock-claude-sonnet",
            type=click.Choice(bedrock_models),
            show_choices=False,
        )
        params = dict(
            # Bedrock uses the "model_id" key for what we're calling "long_model_id"
            model_id=MODEL_ID_TO_LONG_ID[model_id],
            model_kwargs={"temperature": 0.7},
        )
        max_tokens = TOKEN_LIMITS[MODEL_ID_TO_LONG_ID[model_id]]
        model_cost = COST_PER_1K_TOKENS[MODEL_ID_TO_LONG_ID[model_id]]
        cfg = {
            "model_type": model_type,
            "model_id": model_id,
            "model_args": params,
            "token_limit": max_tokens,
            "model_cost": model_cost,
            "input_token_proportion": 0.4,
        }
    else:
        raise ValueError(f"Unknown model type {model_type}")
    with open(model_cfg, "w") as f:
        json.dump(cfg, f, indent=2)
    print(f"Model config written to {model_cfg}")


@llm.command("ls", help="List all of the user-configured models")
def llm_ls(
    all: Annotated[
        bool,
        typer.Option(
            "--all",
            "-a",
            is_flag=True,
            help="List all models, including the default model IDs.",
            click_type=click.Choice(sorted(list(MODEL_TYPE_CONSTRUCTORS.keys()))),
        ),
    ] = False,
):
    import json

    from janus.cli.constants import MODEL_CONFIG_DIR
    from janus.llm.models_info import MODEL_TYPES

    print("\n[green]User-configured models[/green]:")
    for model_cfg in MODEL_CONFIG_DIR.glob("*.json"):
        with open(model_cfg, "r") as f:
            cfg = json.load(f)
        print(f"\t[blue]{model_cfg.stem}[/blue]: [purple]{cfg['model_type']}[/purple]")

    if all:
        print("\n[green]Available model IDs[/green]:")
        for model_id, model_type in MODEL_TYPES.items():
            print(f"\t[blue]{model_id}[/blue]: [purple]{model_type}[/purple]")
